import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib import cm
from genepy.epigenetics import chipseq as chip
from genepy.utils import helper as h

def plotAverageOfSamples(samples, folder="", showAll=False, maxv=None, minv=None):
  res = [] 
  plt.figure()
  plt.ylim(minv,maxv)
  for sample in samples:
    data = pd.read_csv(sample, sep='\t', skiprows=1, header=None, names=['chr', 'start', 'end', 'name', "foldchange","."]+list(range(600)))
    r = data[list(range(600))].mean().tolist()
    res.append(r)
    if showAll:
      sns.lineplot(data=np.array(r), color="#BFBFFF")
  sns.lineplot(data=np.array(res).mean(0))
  if folder:
    plt.savefig(folder+"_averageofsamples.pdf", color="#1F1FFF")
  return res


def pysam_getPeaksAt(peaks, bams, folder='data/seqs/', window=1000, numpeaks=1000, numthreads=8):

  # get pysam data
  # ask for counts only at specific locus based on windows from center+-size from sorted MYC peaks
  # for each counts, do a rolling average (or a convolving of the data) with numpy
  # append to an array
  # return array, normalized
  loaded = {}
  res = {i: np.zeros((len(peaks), window * 2)) for i in bams}
  peaks = peaks.sort_values(by="foldchange", ascending=False).iloc[:numpeaks]
  peaks.chrom = peaks.chrom.astype(str)
  for val in bams:
    loaded.update({val: pysam.AlignmentFile(
            folder + val, 'rb', threads=numthreads)})
  for k, bam in loaded.items():
    for num, (i, val) in enumerate(peaks.iterrows()):
      print(int(num / len(peaks)), end='\r')
      center = int((val['start'] + val['end']) / 2)
      for pileupcolumn in bam.pileup(val['chrom'], start=center - window,
                                  stop=center + window, truncate=True):
        res[k][num][pileupcolumn.pos - (center - window)] = pileupcolumn.n
  fig, ax = plt.subplots(1, len(res))
  for i, (k, val) in enumerate(res.items()):
    sns.heatmap(val, ax=ax[i])
    ax[i].set_title(k.split('.')[0])
  fig.show()
  return res, fig


def bedtools_getPeaksAt(peaks, bams, folder='data/seqs/', window=1000, numpeaks=1000, numthreads=8):
  """
  get pysam data
  ask for counts only at specific locus based on windows from center+-size from sorted MYC peaks
  for each counts, do a rolling average (or a convolving of the data) with numpy
  append to an array
  return array, normalized
  """
  loaded = {}
  center = [int((val['start'] + val['end']) / 2) for k, val in peaks.iterrows()]
  peaks['start'] = [c - window for c in center]
  peaks['end'] = [c + window - 1 for c in center]
  peaks[peaks.columns[:3]].sort_values(by=['chrom', 'start']).to_csv(
            'temp/peaks.bed', sep='\t', index=False, header=False)
  bedpeaks = BedTool('temp/peaks.bed')

  fig, ax = plt.subplots(1, len(bams))
  peakset = peaks["foldchange"].values.argsort()[::-1][:numpeaks]
  for i, val in enumerate(bams):
    coverage = BedTool(folder + val).intersect(bedpeaks).genome_coverage(bga=True, split=True)\
      .intersect(bedpeaks).to_dataframe(names=['chrom', 'start', 'end', 'coverage'])
    cov = np.zeros((len(peaks), window * 2), dtype=int)
    j = 0
    pdb.set_trace()
    for i, (k, val) in enumerate(peaks.iterrows()):
      print(i / len(peaks), end='\r')
      while coverage.iloc[j].start > val.start:
        j -= 1
      while coverage.iloc[j].start < val.end:
        cov[i][coverage.iloc[j].start - val.start:coverage.iloc[j].end - val.start] =\
          coverage.iloc[j].coverage
        j += 1
    sns.heatmap(coverage, ax=ax[i])
    ax[i].set_title(val.split('.')[0])
  fig.show()
  return None, fig


def makeProfiles(matx=[], folder='', matnames=[], title='',
                 name='temp/peaksat.pdf', refpoint="TSS", scale=None,
                 sort=False, withDeeptools=True, cluster=1, vmax=None, vmin=None, overlap=False,
                 legendLoc=None):
  if withDeeptools:
    if not (len(matnames) == 2 and len(matx) == 2):
      raise ValueError('you need two mat.gz files and two names')
    h.createFoldersFor(name)
    cmd = 'computeMatrixOperations relabel -m '
    cmd += matx[0] + ' -o '+matx[0]+' --groupLabels '+matnames[0]
    cmd += ' && computeMatrixOperations relabel -m '
    cmd += matx[1] + ' -o '+matx[1]+' --groupLabels '+matnames[1]
    cmd += ' && computeMatrixOperations rbind -m '
    cmd += matx[0] + ' ' + matx[1] + " -o " + \
                    '.'.join(name.split('.')[:-1]) + ".gz"
    cmd += ' && plotProfile'
    cmd += " --matrixFile " + '.'.join(name.split('.')[:-1]) + ".gz"
    cmd += " --outFileName " + name
    cmd += " --refPointLabel " + refpoint
    if vmax is not None:
      cmd += " -max "+str(vmax)
    if vmin is not None:
      cmd += " -min "+str(vmin)
    if cluster > 1:
      cmd += " --perGroup --kmeans "+str(cluster)
    if legendLoc:
      cmd += " --legendLocation "+legendLoc
    if title:
      cmd += " --plotTitle " + title
    data = subprocess.run(cmd, shell=True, capture_output=True)
    print(data)


def getPeaksAt(peaks, bigwigs, folder='', bigwignames=[], peaknames=[], window=1000, title='', numpeaks=4000, numthreads=8,
               width=5, length=10, torecompute=False, name='temp/peaksat.pdf', refpoint="TSS", scale=None,
               sort=False, withDeeptools=True, onlyProfile=False, cluster=1, vmax=None, vmin=None, overlap=False,
               legendLoc=None):
  """
  get pysam data
  ask for counts only at specific locus based on windows from center+-size from sorted MYC peaks
  for each counts, do a rolling average (or a convolving of the data) with numpy
  append to an array
  return array, normalized
  """
  if withDeeptools:
    if isinstance(peaks, pd.DataFrame):
      peaks = 'peaks.bed '
      peaks.to_csv('peaks.bed', sep='\t', index=False, header=False)
    elif type(peaks) == list:
      pe = ''
      i = 0
      for n, p in enumerate(peaks):
        if 20 < int(os.popen('wc -l ' + p).read().split(' ')[0]):
          pe += p + ' '
        elif len(peaknames) > 0:
          peaknames.pop(n-i)
          i += 1
      peaks = pe
    elif type(peaks) == str:
      peaks += ' '
    else:
      raise ValueError(' we dont know this filetype')
    if type(bigwigs) is list:
      pe = ''
      for val in bigwigs:
        pe += folder + val + ' '
      bigwigs = pe
    else:
      bigwigs = folder + bigwigs + ' '
    h.createFoldersFor(name)
    cmd = ''
    if not os.path.exists('.'.join(name.split('.')[:-1]) + ".gz") or torecompute:
      cmd += "computeMatrix reference-point -S "
      cmd += bigwigs
      cmd += " --referencePoint "+refpoint
      cmd += " --regionsFileName " + peaks
      cmd += " --missingDataAsZero"
      cmd += " --outFileName " + '.'.join(name.split('.')[:-1]) + ".gz"
      cmd += " --upstream " + str(window) + " --downstream " + str(window)
      cmd += " --numberOfProcessors " + str(numthreads) + ' && '
    cmd += "plotHeatmap" if not onlyProfile else 'plotProfile'
    if type(name) is list:
      if not onlyProfile:
        raise ValueError('needs to be set to True, can\'t average heatmaps')
      cmd += " --matrixFile " + '.gz '.join(name) + ".gz"
      if average:
        cmd += "--averageType mean"
    else:
      cmd += " --matrixFile " + '.'.join(name.split('.')[:-1]) + ".gz"
    cmd += " --outFileName " + name
    cmd += " --refPointLabel " + refpoint
    if vmax is not None:
      cmd += " -max "+str(vmax)
    if vmin is not None:
      cmd += " -min "+str(vmin)
    if cluster > 1:
      cmd += " --perGroup --kmeans "+str(cluster)
    if overlap:
      if onlyProfile:
        cmd += " --plotType overlapped_lines"
      else:
        raise ValueError("overlap only works when onlyProfile is set")
    if legendLoc:
      cmd += " --legendLocation "+legendLoc

    if len(peaknames) > 0:
      pe = ''
      for i in peaknames:
        pe += ' ' + i
      cmd += " --regionsLabel" + pe
    if type(bigwigs) is list:
      if len(bigwignames) > 0:
        pe = ''
        for i in bigwignames:
          pe += ' "' + i + '"'
        cmd += " --samplesLabel" + pe
    if title:
      cmd += " --plotTitle '"+title+"'"
    data = subprocess.run(cmd, shell=True, capture_output=True)
    print(data)
  else:
    if 'relative_summit_pos' in peaks.columns:
      center = [int((val['start'] + val['relative_summit_pos']))
                    for k, val in peaks.iterrows()]
    else:
      center = [int((val['start'] + val['end']) / 2)
                    for k, val in peaks.iterrows()]
    pd.set_option('mode.chained_assignment', None)
    peaks['start'] = [c - window for c in center]
    peaks['end'] = [c + window for c in center]
    fig, ax = plt.subplots(1, len(bigwigs), figsize=[
                           width, length], title=title if title else 'Chip Heatmap')
    if sort:
      peaks = peaks.sort_values(by=["foldchange"], ascending=False)
    if numpeaks > len(peaks):
      numpeaks = len(peaks) - 1
    cov = {}
    maxs = []
    for num, bigwig in enumerate(bigwigs):
      bw = pyBigWig.open(folder + bigwig)
      co = np.zeros((numpeaks, window * 2), dtype=int)
      scale = scale[bigwig] if scale is dict else 1
      for i, (k, val) in enumerate(peaks.iloc[:numpeaks].iterrows()):
        try:
          co[i] = np.nan_to_num(bw.values(str(val.chrom), val.start, val.end), 0)
        except RuntimeError as e:
          print(str(val.chrom), val.start, val.end)
          pass
      cov[bigwig] = co
      maxs.append(co.max())
    for num, bigwig in enumerate(bigwigs):
      sns.heatmap(cov[bigwig] * scale, ax=ax[num], vmax=max(maxs), yticklabels=[], cmap=cmaps[num],
                    cbar=True)
      ax[num].set_title(bigwig.split('.')[0])
    fig.subplots_adjust(wspace=0.1)
    fig.show()
    fig.savefig(name)
    return cov, fig


def andrew(groups, merged, annot, enr=None, pvals=None, cols=8, precise=True, title = "sorted clustermap of cobindings clustered", folder="", rangeval=4, okpval=10**-3, size=(20,15),vmax=3, vmin=0):
  if enr is None or pvals is None:
    enr, pvals = chip.enrichment(merged, groups=groups)
  rand = np.random.choice(merged.index,5000)
  subgroups = groups[rand]
  sorting = np.argsort(subgroups)
  redblue = cm.get_cmap('RdBu_r',256)
  subenr = enr.iloc[annot-cols:]
  subenr[subenr>rangeval]=rangeval
  subenr[subenr<-rangeval]=-rangeval
  subenr = subenr/rangeval
  data = []
  #colors = []
  impv = pvals.values
  for i in subgroups[sorting]:
    #colors.append(viridis(i))
    a = redblue((128+(subenr[i]*128)).astype(int)).tolist()
    for j in range(len(a)):
      a[j] = [1.,1.,1.,1.] if impv[j,i] > okpval else a[j]
    data.append(a)
  data = pd.DataFrame(data=data,columns=list(subenr.index),index= rand[sorting])
  #data["clusters"]  = colors
  
  a = np.log2(1.01+merged[merged.columns[cols:annot]].iloc[rand].iloc[sorting].T)
  if not precise:
    for i in set(groups):
      e = a[a.columns[subgroups[sorting]==i]].mean(1)
      e = pd.DataFrame([e for i in range((subgroups[sorting]==i).sum())]).T
      a[a.columns[subgroups[sorting]==i]] = e
  
  fig = sns.clustermap(a, vmin=vmin, vmax=vmax, figsize=size, z_score=0, colors_ratio=0.01, col_cluster=False,col_colors=data, xticklabels=False)
  fig.ax_col_dendrogram.set_visible(False)
  fig.fig.suptitle(title)
  fig.savefig(folder + str(len(set(groups))) + '_clustermap_cobinding_enrichment_andrewplot.pdf')
  plt.show()



import gspread
from oauth2client.service_account import ServiceAccountCredentials

scope = ["https://spreadsheets.google.com/feeds", 'https://www.googleapis.com/auth/spreadsheets',
         "https://www.googleapis.com/auth/drive.file", "https://www.googleapis.com/auth/drive"]

credentials = ServiceAccountCredentials.from_json_keyfile_name('~/.client_secret.json', scope)
client = gspread.authorize(credentials)

spreadsheet = client.open('https://docs.google.com/spreadsheets/d/1XkZypRuOEXzNLxVk9EOHeWRE98Z8_DBvL4PovyM01FE')

with open(file, 'r') as file_obj:
  content = file_obj.read()
  client.import_csv(spreadsheet.id, data=content)



###########################################################
#
# PYDESEQ
#
##################################################################

from __future__ import print_function
import numpy as np
import rpy2.robjects as robjects
from rpy2.robjects import pandas2ri, Formula, numpy2ri
pandas2ri.activate()
import rpy2
from rpy2.robjects.packages import importr
deseq = importr('DESeq2')
from rpy2.robjects.conversion import localconverter
import rpy2.robjects as ro
import sys
'''
Adopted from: https://stackoverflow.com/questions/41821100/running-deseq2-through-rpy2
'''

to_dataframe = robjects.r('function(x) data.frame(x)')


class pyDESeq2:
  '''
  DESeq2 object through rpy2
  input:
  count_matrix: should be a pandas dataframe with each column as count, and a id column for gene id
      example:
      id    sampleA    sampleB
      geneA    5    1
      geneB    4    5
      geneC    1    2
  design_matrix: an design matrix in the form of pandas dataframe, see DESeq2 manual, samplenames as rownames
              treatment
  sampleA1        A
  sampleA2        A
  sampleB1        B
  sampleB2        B
  design_formula: see DESeq2 manual, example: "~ treatment""
  gene_column: column name of gene id columns, exmplae "id"
  '''

  def __init__(self, count_matrix, design_matrix, design_formula, gene_column='gene_id'):
    print("you need to have R installed with the DESeq2 library installed")
    try:
      assert gene_column == count_matrix.columns[0], 'no $gene_column name in 1st column\'s name'
      gene_id = count_matrix[gene_column]
    except AttributeError:
      sys.exit('Wrong Pandas dataframe?')
    print(rpy2.__version__)
    self.deseq_result = None
    self.resLFC = None
    self.comparison = None
    self.normalized_count_matrix = None
    self.gene_column = gene_column
    self.gene_id = count_matrix[self.gene_column]
    with localconverter(ro.default_converter + pandas2ri.converter):
      self.count_matrix = pandas2ri.py2rpy(count_matrix.drop(gene_column, axis=1).astype(int))
      self.design_matrix = pandas2ri.py2rpy(design_matrix.astype(bool))
    self.design_formula = Formula(design_formula)
    self.dds = deseq.DESeqDataSetFromMatrix(countData=self.count_matrix,
                                            colData=self.design_matrix,
                                            design=self.design_formula)

  def run_estimate_size_factors(self, **kwargs):  # OPTIONAL
    """
    args:
      geoMeans: cond*gene matrix
    """
    self.dds = deseq.estimateSizeFactors_DESeqDataSet(self.dds, **kwargs)

  def run_deseq(self, **kwargs):
    self.dds = deseq.DESeq(self.dds, **kwargs)

  def getSizeFactors(self):
    return deseq.sizeFactors_DESeqDataSet(self.dds)

  def setSizeFactors(self, factors):
    val = self.dds.do_slot('colData').do_slot('listData')
    val[2] = ro.vectors.FloatVector(np.array(factors))
    self.dds.do_slot('colData').do_slot_assign('listData', val)

  def get_deseq_result(self, **kwargs):

    self.comparison = deseq.resultsNames(self.dds)

    self.deseq_result = deseq.results(self.dds, **kwargs)
    self.deseq_result = to_dataframe(self.deseq_result)
    with localconverter(ro.default_converter + pandas2ri.converter):
      self.deseq_result = ro.conversion.rpy2py(self.deseq_result)  # back to pandas dataframe
    self.deseq_result[self.gene_column] = self.gene_id.values



import numpy as np


def getDFinfo(df):
  val = df
  print("sums over cell lines! ------mean, var, totmin, meanmin, totmax, meanmax")
  print(val.sum(1).mean(), val.sum(1).var(), val.sum(1).min(), val.mean(1).min(), val.sum(1).idxmin(), val.sum(1).max(), val.mean(1).max(), val.sum(1).idxmax())
  print("sums over features! ------mean, var, totmin, meanmin, totmax, meanmax")
  print(val.sum(0).mean(), val.sum(0).var(), val.sum(0).min(), val.mean(0).min(), val.sum(0).idxmin(), val.sum(0).max(), val.mean(0).max(), val.sum(0).idxmax())
  print("nans!")
  print(np.count_nonzero(np.isnan(val)))


def compare(df1, df2):
  df = pd.concat([df1, df2])
  df = df.reset_index(drop=True)
  df_gpby = df.groupby(list(df.columns))
  idx = [x[0] for x in df_gpby.groups.values() if len(x) == 1]
  df.reindex(idx)
  return df





# Jeremie Kalfon
# for BroadInsitute
# in 2019

from __future__ import print_function
import warnings
from matplotlib import pyplot as plt
from bokeh.palettes import *
from bokeh.plotting import *
from scipy.stats import pearsonr
from genepy.utils import helper as h
import math
import os
import seaborn as sns

# import gseapy
import pandas as pd
import numpy as np
import subprocess
from sklearn.decomposition import PCA, IncrementalPCA
import umap.umap_ as umap
from genepy.utils import plot as hplot


def filterProteinCoding(listofgenes, from_idtype="ensemble_gene_id"):
    """
    Given a list of genes, provide the args where the genes are protein coding genes:

    This functtion will use a file in taiga, you need taigapy installed

    Args:
    -----
      listofgenes: list of genes
      from_idtype: one of "symbol","uniprot_ids","pubmed_id","ensembl_gene_id","entrez_id","name", the gene name format

    Returns:
    -------
      the args where the genes are protein coding
    """
    tokeep = []
    b = 0
    gene_mapping = h.generateGeneNames()

    for i, val in enumerate(listofgenes):
        if from_idtype == "ensemble_gene_id":
            val = val.split(".")[0]
        elif from_idtype == "hgnc_id":
            val = "HGNC:" + str(val)
        a = gene_mapping["locus_group"][gene_mapping[from_idtype] == val].values
        if len(a) > 0:
            if a[0] == "protein-coding gene":
                tokeep.append(i)
        else:
            b += 1
    print(str(b))
    return tokeep


def convertGenes(
    listofgenes,
    from_idtype="ensembl_gene_id",
    to_idtype="hgnc_symbol",
    useCache=True,
    ensemble_server="http://nov2020.archive.ensembl.org/biomart",
):
    """
    Given a list of genes, will convert to different ID type

    Args:
    -----
      listofgenes: list of genes
      from_idtype: one of "ensembl_gene_id","clone_based_ensembl_gene","hgnc_symbol","gene_biotype","entrezgene_id", the gene name format
      to_idtype: one of "ensembl_gene_id","clone_based_ensembl_gene","hgnc_symbol","gene_biotype","entrezgene_id", the gene name format

    Returns:
    -------
      1: the new names for each genes that were matched else the same name
      2: the names of genes that could not be matched
    """

    gene_mapping = h.generateGeneNames(
        useCache=useCache, ensemble_server=ensemble_server
    )

    not_parsed = []
    renamed = []
    b = 0
    to = {}

    for _, val in gene_mapping.iterrows():
        to[val[from_idtype]] = val[to_idtype]

    for _, val in enumerate(listofgenes):
        if from_idtype == "ensembl_gene_id":
            val = val.split(".")[0]
        elif from_idtype == "entrezgene_id":
            pass
        try:
            a = to[val]
            renamed.append(a)
        except KeyError:
            b += 1
            not_parsed.append(val)
            renamed.append(val)
    print(str(b) + " count not be parsed... we don't have all genes already")
    return (renamed, not_parsed)


async def getSpikeInControlScales(
    refgenome,
    fastq=None,
    fastQfolder="",
    mapper="bwa",
    pairedEnd=False,
    cores=1,
    pathtosam="samtools",
    pathtotrim_galore="trim_galore",
    pathtobwa="bwa",
    totrim=True,
    tomap=True,
    tofilter=True,
    results="res/",
    toremove=False,
):
    """
    Will extract the spikeInControls from a fastq file (usefull for, let say ChIPseq data with spike ins)

    Count based sequencing data is not absolute and will be normalized as each sample will be sequenced at a specific depth.
    To figure out what was the actual sample concentration, we use Spike In control
    You should have FastQfolder/[NAME].fastq & BigWigFolder/[NAME].bw with NAME being the same for the same samples


    Args:
    -----
      refgenome: str the file path to the indexed reference genome
      FastQfolder: str the folder path where the fastq files are stored (should be named the same as files in BigWigFolder)
      BigWigFolder: str the folder path where the bigwig files are stored (should be named the same as files in FastQfolder)
      mapper: str flag to 'bwa', ...
      pairedEnd: Bool flat to true for paired end sequences. if true, You should have FastQfolder/[NAME]_1|2.fastq

    Returns:
    --------
      dict(file,float) the scaling factor dict

    """
    if len(fastQfolder) > 0:
        print("using all files from folder")
        fastqs = os.listdir(fastQfolder)
        fastqs = [i for i in fastqs if ".fq.gz" == i[-6:] or ".fastq.gz" == i[-9:]]
        fastqs.sort()
        if pairedEnd and (tomap or totrim):
            print("need to be name_*1, name_*2")
            fastqs = [i for i in h.grouped(fastqs, 2)]
    elif fastq is None:
        raise ValueError("you need input files")
    else:
        if type(fastq) is list:
            print("your files need to be all in the same folder")
            fastQfolder = "/".join(fastq[0].split("/")[:-1]) + "/"
            if not totrim and not tomap:
                fastqs = [f.split("/")[-1] for f in fastq]
            else:
                print("need to be name_*1, name_*2")
                fastqs = [
                    [f[0].split("/")[-1], f[1].split("/")[-1]]
                    for f in h.grouped(fastq, 2)
                ]
        else:
            fastQfolder = "/".join(fastq.split("/")[:-1]) + "/"
            fastqs = [fastq.split("/")[-1]]
    print(fastqs)
    if not totrim:
        print("you need to have your files in the " + results + " folder")
    if totrim and tomap:
        print("\n\ntrimming\n\n")
        if pairedEnd:
            cmds = []
            rm = []
            for file in fastqs:
                cmd = (
                    pathtotrim_galore
                    + " --paired --fastqc --gzip "
                    + fastQfolder
                    + file[0]
                    + " "
                    + fastQfolder
                    + file[1]
                    + " -o "
                    + results
                )
                if toremove:
                    rm.append(
                        "rm " + fastQfolder + file[0] + " " + fastQfolder + file[1]
                    )
                cmds.append(cmd)
            print(cmds)
            h.parrun(cmds, cores, add=rm)
            fastqs = [
                [
                    file[0].split(".")[0] + "_val_1.fq.gz",
                    file[1].split(".")[0] + "_val_2.fq.gz",
                ]
                for file in fastqs
            ]
    if tomap:
        print("\n\nmapping\n\n")
        if pairedEnd:
            cmds = []
            rm = []
            for file in fastqs:
                cmd = (
                    pathtobwa
                    + " mem "
                    + refgenome
                    + " "
                    + results
                    + file[0]
                    + " "
                    + results
                    + file[1]
                    + " | "
                    + pathtosam
                    + " sort - -o "
                    + results
                    + file[0].split(".")[0]
                    + ".sorted.bam"
                )
                if toremove:
                    rm.append("rm " + results + file[0] + " " + results + file[1])
                cmds.append(cmd)
        h.parrun(cmds, cores, add=rm)
        fastqs = [file[0].split(".")[0] + ".sorted.bam" for file in fastqs]

    if tofilter:
        print("\n\nfiltering\n\n")
        cmds = []
        rm = []
        h.parrun(
            [
                pathtosam + " index " + results + file.split(".")[0] + ".sorted.bam"
                for file in fastqs
            ],
            cores,
        )
        h.parrun(
            [
                pathtosam
                + " flagstat "
                + results
                + file.split(".")[0]
                + ".sorted.bam > "
                + results
                + file.split(".")[0]
                + ".sorted.bam.flagstat"
                for file in fastqs
            ],
            cores,
        )
        h.parrun(
            [
                pathtosam
                + " idxstats "
                + results
                + file.split(".")[0]
                + ".sorted.bam > "
                + results
                + file.split(".")[0]
                + ".sorted.bam.idxstat"
                for file in fastqs
            ],
            cores,
        )
        fastqs = [file.split(".")[0] + ".sorted.bam" for file in fastqs]
    else:
        print("files need to be named: NAME.sorted.bam")
        fastqs = [file for file in fastqs if ".sorted.bam" == file[-11:]]
    mapped = {}
    norm = {}
    unique_mapped = {}
    print("\n\ncounting\n\n")
    for file in fastqs:
        mapped[file.split(".")[0]] = int(
            os.popen(
                pathtosam
                + " view -c -F 0x004 -F 0x0008 -f 0x001 -F 0x0400 -q 1 "
                + results
                + file
                + " -@ "
                + str(cores)
            )
            .read()
            .split("\n")[0]
        )
        # unique_mapped[file.split('.')[0]] = int(re.findall("Mapped reads: (\d+)", os.popen('bamtools stats -in '+results +
        #                                             file + '.sorted.bam').read())[0])
    nbmapped = np.array([i for i in mapped.values()])
    nbmapped = np.sort(nbmapped)[0] / nbmapped.astype(float)
    for i, val in enumerate(mapped.keys()):
        norm[val] = nbmapped[i]
    return (
        norm,
        mapped,
    )  # unique_mapped


async def GSEAonExperiments(
    data,
    experiments,
    res={},
    savename="",
    scaling=[],
    geneset="GO_Biological_Process_2015",
    cores=8,
    cleanfunc=lambda i: i.split("(GO")[0],
):
    """

    Will run GSEA on a set of experiment

    Args:
    -----
      data: a pandas.df rows: gene counts; columns: [experimentA_X,..., experimentD_X..., control_X] where X is the replicate number
      experiments: a list of experiment names (here experimentA,.. experimentD)
      scaling: a dict(experiment:(mean,std)) of scaling factors and their associated standard error for each experiments
      res: you can provide a dict containing results from
      savename: if you want to save the plots as pdfs, provides a location/name
      geneset: the geneset to run it on. (can be a filepath to your own geneset)
      cores: to run GSEA on
      cleanfunc: a func applied to the names of the gene sets to change it in some way (often to make it more readable)
    Returns
    -------
      plots the results
      1: returns a matrix with the enrichment for each term for each experiment
      2: returns a dict(experiment:pd.df) with dataframe being the output of GSEA (with pvalues etc..) for each experiments
    """
    for i, val in enumerate(experiments):
        print(val)
        totest = data[[v for v in data.columns[:-1] if val + "-" in v or "AAVS1" in v]]
        cls = ["Condition" if val + "-" in v else "DMSO" for v in totest.columns]
        if scaling:
            if abs(scaling[val.split("_")[1]][0]) > scaling[val.split("_")[1]][1]:
                print("rescaling this one")
                cols = [i for i in totest.columns if val + "-" in i]
                totest[cols] = totest[cols] * (2 ** scaling[val.split("_")[1]][0])
        if val in res:
            print(val + " is already in set")
            continue
        res[val] = gseapy.gsea(
            data=totest, gene_sets=geneset, cls=cls, no_plot=False, processes=cores
        )
        res[val].res2d["Term"] = [i for i in res[val].res2d.index]
        for i, v in res.items():
            res[i].res2d["Term"] = [cleanfunc(i) for i in v.res2d["Term"]]
        plt.figure(i)
        sns.barplot(
            data=res[val].res2d.iloc[:25], x="es", y="Term", hue_order="geneset_size"
        ).set_title(val)
    a = set()
    for k, val in res.items():
        a.update(set(val.res2d.Term))
    a = {i: [0] * len(res) for i in a}
    for n, (k, val) in enumerate(res.items()):
        for i, v in val.res2d.iterrows():
            a[v.Term][n] = v.es
    pres = pd.DataFrame(a, index=res.keys())
    a = sns.clustermap(
        figsize=(25, 20),
        data=res,
        vmin=-1,
        vmax=1,
        yticklabels=res.index,
        cmap=plt.cm.RdYlBu,
    )
    b = sns.clustermap(-res.T.corr(), cmap=plt.cm.RdYlBu, vmin=-1, vmax=1)
    if savename:
        res.to_csv(savename + ".csv")
        a.savefig(savename + "_genesets.pdf")
        b.savefig(savename + "_correlation.pdf")
    return pres, res


def runERCC(
    ERCC,
    experiments,
    featurename="Feature",
    issingle=False,
    dilution=1 / 100,
    name="RNPv2",
    spikevol=1,
    control="AAVS1",
    fdr=0.1,
    totalrnamass=0.5,
):
    """
    Runs the ERCC dashboard Rpackage on your notebook

    you will need to run this function from ipython and to have the R package erccdashboard installed

    Args:
    ----
      ERCC: a pandas.df rows: ERCC counts columns: [experimentA_X,..., experimentD_X..., control_X] where X is the replicate number
      experiments: a list of experiment names (here experimentA,.. experimentD)
      featurename: columns where the ERCC pseudo gene names are stored
      issingle: ERCC parameters to choose between Single and RatioPair
      dilution: ERCC dilution parameter
      name: the name of the experiment set
      spikevol: ERCC spikevol parameter
      control: the control name (here control)
      fdr: ERCC fdr parameter
      totalrnamass: ERCC totalrnamass parameter

    Returns:
    -------
      a dict(experimentName:(val, ste)) a dict containing the scaling factor and its standard error for each experiment

    Raises:
    ------
      RuntimeError: if you are not on ipython
    """
    try:
        ipython = get_ipython()
    except:
        raise RuntimeError("you need to be on ipython")
    print("you need to have R installed with the erccdashboard library installed")
    ipython.magic("load_ext rpy2.ipython")
    ipython.magic("R library('erccdashboard')")
    # "count" for RNA-Seq data, "array" for microarray data
    ipython.magic("R datType = 'count'")
    # flag to indicate if input expression measures are already normalized, default is FALSE
    ipython.magic("R isNorm = F")
    # user defined filename prefix for results files
    ipython.magic("R -i name filenameRoot = name")
    # name for sample 2 in the experiment
    ipython.magic("R -i control sample2Name = control")
    # name of ERCC mixture design, "RatioPair" is default
    ipython.magic("R -i issingle erccmix <- if(issingle) 'Single' else 'RatioPair'")
    # dilution factor used for Ambion spike-in mixtures
    ipython.magic("R -i dilution erccdilution = dilution")
    # volume (in microliters) of diluted spike-in mixture added to total RNA mass
    ipython.magic("R -i spikevol spikeVol = spikevol")
    # user defined false discovery rate (FDR), default is 0.05
    ipython.magic("R -i fdr choseFDR = fdr")
    ipython.magic("R exDat = ''")
    ipython.magic("R -i totalrnamass totalRNAmass <- totalrnamass")

    cols = list(ERCC.columns)
    cols.sort()
    res = {}
    for val in experiments:
        d = {}
        e = 0
        c = 0
        d.update({featurename: "Feature"})
        cont = {}
        for i in cols:
            if val + "-" in i:
                e += 1
                d.update({i: val.split("_")[-1] + "_" + str(e)})
            if control + "-" in i:
                c += 1
                cont.update({i: control + "_" + str(c)})
        d.update(cont)
        a = ERCC[list(d.keys())].rename(columns=d)
        a.to_csv("/tmp/ERCC_estimation.csv", index=None)
        val = val.split("_")[-1]

        ipython.magic("R -i val print(val)")
        ipython.magic("R print(sample2Name)")
        ipython.magic("R a <- read.csv('/tmp/ERCC_estimation.csv')")
        ipython.magic("R print(head(a))")

        try:
            ipython.magic(
                "R -i val exDat = initDat(datType=datType, isNorm=isNorm, exTable=a,\
                            filenameRoot=filenameRoot, sample1Name=val,\
                            sample2Name=sample2Name, erccmix=erccmix,\
                            erccdilution=erccdilution, spikeVol=spikeVol,\
                            totalRNAmass=totalRNAmass, choseFDR=choseFDR)"
            )
            ipython.magic("R exDat = est_r_m(exDat)")
            ipython.magic("R exDat = dynRangePlot(exDat)")

        except Warning:
            print("failed for " + val)
            continue
        except:
            print("worked for " + val)

        ipython.magic("R print(summary(exDat))")
        ipython.magic("R grid.arrange(exDat$Figures$dynRangePlot)")
        ipython.magic("R grid.arrange(exDat$Figures$r_mPlot)")
        ipython.magic("R grid.arrange(exDat$Figures$rangeResidPlot)")

        ipython.magic("R -o rm rm <- exDat$Results$r_m.res$r_m.mn")
        ipython.magic("R -o se se <- exDat$Results$r_m.res$r_m.mnse")
        ipython.magic("R write.csv(c(rm,se), file = '/tmp/genepy_ercc.csv')")
        ipython.magic("R print(se,rm)")
        l = h.fileToList("/tmp/genepy_ercc.csv")
        res[val] = (float(l[1][4:]), float(l[2][4:]))
    for i, v in res.items():
        if abs(v[0]) > v[1]:
            print(i, v[0])
    return res


def mergeSplicingVariants(df, defined="."):
    df = df.sort_index()
    foundpoint = False
    # pdb.set_trace()
    torename = {}
    todrop = []
    for i, v in enumerate(df.index.tolist()):
        h.showcount(i, len(df))
        if foundpoint:
            if foundpoint in v:
                tomerge.append(v)
            else:
                if foundpoint not in df.index:
                    if len(tomerge) > 1:
                        # print("merging "+str(len(tomerge)))
                        df.loc[foundpoint] = df.loc[tomerge].sum()
                        todrop.extend(tomerge)
                    else:
                        torename.update({tomerge[0]: foundpoint})
                else:
                    todrop.extend(tomerge)
                    tomerge.append(foundpoint)
                    df.loc[foundpoint] = df.loc[tomerge].sum()
                foundpoint = False
        elif defined in v:
            foundpoint = v.split(defined)[0]
            tomerge = [v]
    if len(torename) > 0:
        df = df.rename(index=torename)
    df = df.drop(index=todrop)
    return df


def readFromSlamdunk(
    loc="res/count/",
    flag_var=100,
    convertTo="hgnc_symbol",
    minvar_toremove=0,
    mincount_toremove=5,
    verbose=True,
):

    """
    Given a list of genes, provide the args where the genes are protein coding genes (or given biotype):

    Args:
    -----
      listofgenes: list of genes
      from_idtype: one of "ensemble_gene_id","clone_based_ensembl_gene","hgnc_symbol","gene_biotype","entrezgene_id", the gene name format
      to_idtype: one of "ensemble_gene_id","clone_based_ensembl_gene","hgnc_symbol","gene_biotype","entrezgene_id", the gene name format
      gene_biotype: gene/transcript biotype
      verbose: print total counts and t to c converted counts at MYC

    Returns:
    -------
      1: the new names for each genes that were matched else the same name
      2: the names of genes that could not be matched
    """
    # sorted files
    files = sorted(os.listdir(loc))
    files = [file for file in files if file.endswith(".tsv")]
    data = {}
    for file in files:
        data[file.split("/")[-1].split(".")[0]] = pd.read_csv(
            loc + file, sep="\t", comment="#", header=0
        )

    prev = -2
    print("found " + str(len(data)) + " files:" + str(data.keys()))

    for k, val in data.items():
        if len(set(val.Name)) != prev and prev != -2:
            raise ValueError("we do not have the same number of genes in each file")
        prev = len(set(val.Name))

    # make dict for each unique gene of list of 0s per sample (can be multiple regions)
    readcounts = {i: [0] * len(data) for i in val.Name.unique()}
    tccounts = {i: [0] * len(data) for i in val.Name.unique()}

    for n, (_, val) in enumerate(data.items()):
        print(_.split("_tcount")[0])
        # make df rows ordered by gene name
        val = val.sort_values(by="Name")
        j = 0
        # get ReadCount at first row
        readcount = [val.iloc[0].ReadCount]
        # get TcReadCount at first row
        tccount = [val.iloc[0].TcReadCount]
        # get row Name
        prevname = val.iloc[0].Name

        # repeat for all rows
        for _, v in val.iloc[1:].iterrows():
            if v.Name == 4609 and verbose:
                # MYC region (for QC purposes)
                print(
                    "MYC (readcounts, tccounts): {}, {}".format(
                        v.ReadCount, v.TcReadCount
                    )
                )
            if v.Name == prevname:
                # add counts to rows with the same name
                readcount.append(v.ReadCount)
                tccount.append(v.TcReadCount)
            else:
                # sum read counts in rows with the same name
                readcounts[prevname][n] = np.sum(readcount)
                tccounts[prevname][n] = np.sum(tccount)
                if np.var(readcount) > flag_var:
                    print("pb with " + str(v.Name))
                # new gene name for region
                prevname = v.Name
                j += 1
                if verbose:
                    print(j, end="\r")
                # get read count for new region
                readcount = [v.ReadCount]
                tccount = [v.TcReadCount]

    files = [*data]
    readcounts = pd.DataFrame(
        data=readcounts, columns=val.Name.unique(), index=data.keys()
    ).T
    tccounts = pd.DataFrame(
        data=tccounts, columns=val.Name.unique(), index=data.keys()
    ).T

    # convert to gene symbols
    if convertTo:
        names, _ = convertGenes(
            readcounts.index.tolist(),
            from_idtype="entrezgene_id",
            to_idtype="hgnc_symbol",
        )
        readcounts.index = names
        names, _ = convertGenes(
            tccounts.index.tolist(),
            from_idtype="entrezgene_id",
            to_idtype="hgnc_symbol",
        )
        tccounts.index = names
    # remove regions/genes with variance of 0 across samples
    nottodrop = np.argwhere(tccounts.values.var(1) >= minvar_toremove).ravel()
    tccounts = tccounts.iloc[nottodrop]
    readcounts = readcounts.iloc[nottodrop]
    # remove regions/genes with very low counts
    nottodrop = np.argwhere(readcounts.values.max(1) >= mincount_toremove).ravel()
    tccounts = tccounts.iloc[nottodrop]
    readcounts = readcounts.iloc[nottodrop]
    return readcounts, tccounts


def DESeqSamples(
    data,
    experiments,
    scaling=None,
    keep=True,
    rescaling=None,
    results={},
    controlcontain="RNP_AAVS1",
    spikecontrolscontain="ERCC-",
    threshforscaling=2,
):
    """
    Args:
      data (pd.dataframe): the expected counts with gene names as index and *Experiments[*rep], controlcontain[*rep] as columns
      experiments (list): the names of the different conditions (should be part of your conditions column names)
      scaling (dict(condition/control: float)|list|bool): if boolean, looks for ERCC genes to scale
        if dict, will manually scale each columns condition, control to the provided value
        if list, will use that list as a list of housekeeping genes to use as a reference for rescaling
      keep ():
      rescaling ():
      controlcontain (str): the name of th control (should be part of your control column names).
      spikecontrolscontain (str): the name of the ERCC genes. defaul to ERCC- (for ERCC-1, ERCC-2...)
      threshforscaling (float): minimum scaling value to scale the data for (when scaling is dict)

    Returns:
    ------
      (pd.dataframe): deseq results

    """
    from genepy.rna import pyDESeq2

    if "gene_id" not in list(data.columns):
        print("using index as gene_id")
        data["gene_id"] = data.index
    warnings.simplefilter("ignore")
    if type(controlcontain) is str:
        controlcontain = [controlcontain] * len(experiments)
    for j, val in enumerate(experiments):
        print(val)
        cond = [1 if val + "-" in i else 0 for i in data.columns[:-1]]
        contr = [1 if controlcontain[j] in i else 0 for i in data.columns[:-1]]
        design = pd.DataFrame(
            index=data.columns[:-1],
            columns=["DMSO", "Target"],
            data=np.array([contr, cond]).T,
        )
        design.index = design.index.astype(str).str.replace("-", ".")
        deseq = pyDESeq2.pyDESeq2(
            count_matrix=data,
            design_matrix=design,
            design_formula="~Target",
            gene_column="gene_id",
        )
        if type(scaling) is bool:
            print("  scaling using ERCC")
            if scaling:
                deseq.run_estimate_size_factors(
                    controlGenes=data.gene_id.str.contains(spikecontrolscontain)
                )
        elif type(scaling) is list or type(scaling) is set:
            print("  scaling using a gene set")
            deseq.run_estimate_size_factors(controlGenes=data.gene_id.isin(scaling))
        elif type(scaling) is dict:
            if val in scaling:
                print("  auto scaling from ERCCdashboard mean/std values")
                if abs(scaling[val][0]) > threshforscaling * scaling[val][1]:
                    print("  estimating sizeFactors for this one")
                    deseq.run_estimate_size_factors(
                        controlGenes=data.gene_id.str.contains(spikecontrolscontain)
                    )
                    if rescaling is not None:
                        if val in rescaling:
                            sizeFact = deseq.getSizeFactors()
                            sizeFact[np.where(cond)[0]] *= rescaling[val.split("_")[1]]
                            deseq.setSizeFactors(sizeFact)
            else:
                print("  not in scaling dict")
        elif val in results and keep:
            continue
        deseq.run_deseq()
        deseq.get_deseq_result()
        r = deseq.deseq_result
        r.pvalue = np.nan_to_num(np.array(r.pvalue), 1)
        r.log2FoldChange = np.nan_to_num(np.array(r.log2FoldChange), 0)
        results[val] = r
        print("____________________________________________")
    return results


async def gsva(data, geneset_file, method="ssgsea", recompute=True):
    print("you need to have R installed with GSVA and GSEABase library installed")
    pathtogenepy = os.path.dirname(os.path.abspath(__file__))
    if (
        not recompute
        and os.path.exists("/tmp/data_genepyhelper_gsva.csv")
        and os.path.exists("/tmp/res_genepy_ssGSEA.tsv")
    ):
        print("trying to bypass computing...")
        v = pd.read_csv("/tmp/data_genepyhelper_gsva.csv", index_col=0)
        if v.shape[0] != data.shape[0] or v.shape[1] != data.shape[1]:
            print("WARNING: recompute to false but not the same df for sure")
        return pd.read_csv("/tmp/res_genepy_ssGSEA.tsv", sep="\t")
    data.to_csv("/tmp/data_genepyhelper_gsva.csv")
    cmd = (
        "Rscript "
        + pathtogenepy
        + "/ssGSEA.R /tmp/data_genepyhelper_gsva.csv "
        + geneset_file
        + " "
        + method
    )
    res = subprocess.run(cmd, shell=True, capture_output=True)
    if res.returncode != 0:
        raise ValueError("issue with the command: " + str(res))
    print(res)
    res = pd.read_csv("/tmp/res_genepy_ssGSEA.tsv", sep="\t")
    return res


def filterRNAfromQC(
    rnaqc,
    folder="tempRNAQCplot/",
    plot=True,
    qant1=0.07,
    qant3=0.93,
    thresholds={},
    num_cols=10,
    figsize=(10, 0.2),
):
    thresh = {
        "minmapping": 0.8,  # Mapping Rate
        "minendmapping": 0.75,
        "minefficiency": 0.6,  # Expression Profiling Efficiency
        "maxendmismatch": 0.025,  # Base Mismatch end wise
        "maxmismatch": 0.02,  # Base Mismatch
        "minhighqual": 0.6,  # High Quality Rate
        "minexon": 0.6,  # Exonic Rate
        "maxambiguous": 0.2,  # Ambiguous Alignment Rate
        "maxsplits": 0.1,  # Avg. Splits per Read
        "maxalt": 0.65,  # Alternative Alignments rate
        "maxchim": 0.3,  # Chimeric Alignment Rate
        "minreads": 20000000,
        "minlength": 80,  # Read Length
        "maxgenes": 35000,
        "mingenes": 10000,
    }
    thresh.update(thresholds)

    qcs = rnaqc.T
    tot = []
    a = qcs[
        (qcs["Mapping Rate"] < thresh["minmapping"])
        | (qcs["Base Mismatch"] > thresh["maxmismatch"])
        | (qcs["End 1 Mapping Rate"] < thresh["minendmapping"])
        | (qcs["End 2 Mapping Rate"] < thresh["minendmapping"])
        | (qcs["End 1 Mismatch Rate"] > thresh["maxendmismatch"])
        | (qcs["End 2 Mismatch Rate"] > thresh["maxendmismatch"])
        | (qcs["Expression Profiling Efficiency"] < thresh["minefficiency"])
        | (qcs["High Quality Rate"] < thresh["minhighqual"])
        | (qcs["Exonic Rate"] < thresh["minexon"])
        | (qcs["Ambiguous Alignment Rate"] > thresh["maxambiguous"])
        | (qcs["Avg. Splits per Read"] < thresh["maxsplits"])
        | (qcs["Alternative Alignments"] > thresh["maxalt"] * qcs["Total Reads"])
        | (qcs["Chimeric Alignment Rate"] > thresh["maxchim"])
        | (qcs["Total Reads"] < thresh["minreads"])
        | (qcs["Read Length"] < thresh["minlength"])
        | (thresh["maxgenes"] < qcs["Genes Detected"])
        | (qcs["Genes Detected"] < thresh["mingenes"])
    ].index.tolist()

    tot.append(
        [
            1
            if i in qcs[(qcs["Mapping Rate"] < thresh["minmapping"])].index.tolist()
            else 0
            for i in a
        ]
    )
    tot.append(
        [
            1
            if i in qcs[(qcs["Base Mismatch"] > thresh["maxmismatch"])].index.tolist()
            else 0
            for i in a
        ]
    )
    tot.append(
        [
            1
            if i
            in qcs[(qcs["End 1 Mapping Rate"] < thresh["minendmapping"])].index.tolist()
            else 0
            for i in a
        ]
    )
    tot.append(
        [
            1
            if i
            in qcs[(qcs["End 2 Mapping Rate"] < thresh["minendmapping"])].index.tolist()
            else 0
            for i in a
        ]
    )
    tot.append(
        [
            1
            if i
            in qcs[
                (qcs["End 1 Mismatch Rate"] > thresh["maxendmismatch"])
            ].index.tolist()
            else 0
            for i in a
        ]
    )
    tot.append(
        [
            1
            if i
            in qcs[
                (qcs["End 2 Mismatch Rate"] > thresh["maxendmismatch"])
            ].index.tolist()
            else 0
            for i in a
        ]
    )
    tot.append(
        [
            1
            if i
            in qcs[
                (qcs["Expression Profiling Efficiency"] < thresh["minefficiency"])
            ].index.tolist()
            else 0
            for i in a
        ]
    )
    tot.append(
        [
            1
            if i
            in qcs[(qcs["High Quality Rate"] < thresh["minhighqual"])].index.tolist()
            else 0
            for i in a
        ]
    )
    tot.append(
        [
            1
            if i in qcs[(qcs["Exonic Rate"] < thresh["minexon"])].index.tolist()
            else 0
            for i in a
        ]
    )
    tot.append(
        [
            1
            if i
            in qcs[
                (qcs["Ambiguous Alignment Rate"] > thresh["maxambiguous"])
            ].index.tolist()
            else 0
            for i in a
        ]
    )
    tot.append(
        [
            1
            if i
            in qcs[(qcs["Avg. Splits per Read"] < thresh["maxsplits"])].index.tolist()
            else 0
            for i in a
        ]
    )
    tot.append(
        [
            1
            if i
            in qcs[
                (qcs["Alternative Alignments"] > thresh["maxalt"] * qcs["Total Reads"])
            ].index.tolist()
            else 0
            for i in a
        ]
    )
    tot.append(
        [
            1
            if i
            in qcs[(qcs["Chimeric Alignment Rate"] > thresh["maxchim"])].index.tolist()
            else 0
            for i in a
        ]
    )
    tot.append(
        [
            1
            if i in qcs[(qcs["Total Reads"] < thresh["minreads"])].index.tolist()
            else 0
            for i in a
        ]
    )
    tot.append(
        [
            1
            if i in qcs[(qcs["Read Length"] < thresh["minlength"])].index.tolist()
            else 0
            for i in a
        ]
    )
    tot.append(
        [
            1
            if i in qcs[(thresh["maxgenes"] < qcs["Genes Detected"])].index.tolist()
            else 0
            for i in a
        ]
    )
    tot.append(
        [
            1
            if i in qcs[(qcs["Genes Detected"] < thresh["mingenes"])].index.tolist()
            else 0
            for i in a
        ]
    )

    res = pd.DataFrame(
        index=a,
        columns=[
            "Mapping Rate",
            "Base Mismatch",
            "End 1 Mapping Rate",
            "End 2 Mapping Rate",
            "End 1 Mismatch Rate",
            "End 2 Mismatch Rate",
            "Expression Profiling Efficiency",
            "High Quality Rate",
            "Exonic Rate",
            "Ambiguous Alignment Efficiency",
            "Avg. Splits per Read",
            "Alternative Alignments",
            "Chimeric Alignment Rate",
            "Total Reads",
            "Read Length",
            "Min Genes Detected",
            "Max Genes Detected",
        ],
        data=np.array(tot).astype(bool).T,
    )

    print(a)
    if len(res) > 0:
        h.createFoldersFor(folder)
        res.to_csv(folder + "_qc_results.csv")
        if plot:
            _, ax = plt.subplots(figsize=(figsize[0], math.ceil(len(res) * figsize[1])))
            plot = sns.heatmap(res, xticklabels=True, yticklabels=True, cbar=False)
            plt.yticks(rotation=0)
            plt.show()
            plot.get_figure().savefig(folder + "failed_qc.pdf")

            num_rows = math.ceil(len(rnaqc) / num_cols)
            _, axes = plt.subplots(num_rows, num_cols, figsize=(20, num_rows * 2))
            for val_idx, val in enumerate(rnaqc.index):
                ax = axes.flatten()[val_idx]
                qc = rnaqc.loc[val]
                sns.violinplot(y=qc, ax=ax)
                q1 = qc.quantile(qant1)
                q3 = qc.quantile(qant3)
                outlier_top_lim = q3 + 1.5 * (q3 - q1)
                outlier_bottom_lim = q1 - 1.5 * (q3 - q1)
                for k, v in qc[
                    (qc < outlier_bottom_lim) | (qc > outlier_top_lim)
                ].iteritems():
                    ax.text(
                        0.05,
                        v,
                        k,
                        ha="left",
                        va="center",
                        color="red" if k in a else "black",
                    )
            plt.tight_layout()
            plt.show()
            plt.savefig("{}/qc_metrics.pdf".format(folder), bbox_inches="tight")
    return res


def getDifferencesFromCorrelations(df1, df2, minsimi=0.99999999999999):
    res = []
    overlap = set(df1.columns) & set(df2.columns)
    print(str(len(overlap)) + " overlap")
    df1 = df1[overlap].copy()
    df2 = df2[overlap].copy()
    for k, val in df1.iterrows():
        if k in df2.index:
            corr = np.corrcoef(df1.loc[k], df2.loc[k])
            if corr[0, 1] < minsimi:
                res.append((k, corr[0, 1]))
            else:
                print(k + " not in second df")
    print("found " + str(len(res)) + " samples that did not match")
    return res


def rnaseqcorrelation(cn, rna, ax=None, name=None):
    """
    correlates gene copy number matrix to an expression count matrix

    Shows the correlation plots.
    Gene names should be thee same ones, sample names as welll
    """
    a = set(cn.columns) & set(rna.columns)
    ind = set(cn.index) & set(rna.index)
    re = rna.loc[ind]
    ce = cn.loc[ind]
    print(len(ind), len(a))
    corr = np.array([pearsonr(ce[j], re[j])[0] for j in a])
    # corr = pd.DataFrame(data=corr, columns=[name if name is not None else "data"])
    print(np.mean(corr), len(corr))
    sns.kdeplot(corr, ax=ax) if ax is not None else sns.kdeplot(corr)


def findMissAnnotatedReplicates(repprofiles, goodprofile, names, exactMatch=True):
    """
    from a new rnaseq profile on replicate level and a good rnaseq profile on sample level

    will if some replicates are missanotated based on correlation.

    Returns:
    -------
        notindataset: list[str] replicates not in the good dataset
        missannotated: dict(str: tuple(str,str)).  dict containing replicates that are missanotated: for each, gives a tuple (old annotation, right annotation)
    """
    notindataset = []
    missannotated = {}
    unmatched = {}
    if exactMatch:
        res = findClosestMatching(repprofiles, goodprofile)
        for val in repprofiles.index.tolist():
            if val not in res:
                notindataset.append(val)
            elif val not in names:
                unmatched.update({val: res[val]})
            elif res[val] != names[val]:
                missannotated.update({val: (names[val], res[val])})
        return notindataset, missannotated, unmatched
    else:
        corr, closest = findClosestMatching(repprofiles, goodprofile, returncorr=True)
        for k, v in corr.iterrows():
            print(k, v.mean())
            try:
                if v[names[k]] < 0.75:
                    print(v[[closest[k], names[k]]])
            except:
                a = np.argsort(v.values)[-5:]
                if v.values[a[-1]] > 0.8:
                    print(names[k], corr.columns[a], v.values[a])


def findClosestMatching(repprofiles, goodprofile, closest=False, returncorr=False):
    """
    will find what replicate matches best what known profile using numpy's corrcoef

    Args:
    -----
      repprofiles: dataframe the new  expression profile to test against: dfs should be SAMPLESxGENE
      goodprofile: dataframe the known- expression profile
      closest: bool whether to rerturn the closest matching or just the one that matches perfectly, if any
      returncorr: bool to return the full corelation matrix

    Returns:
    --------
      match: dict(id:id) listing samples that are the closest for all samples
      corr: dataframe of correlations if requested

    """
    match = {}
    a = set(repprofiles.columns) & set(goodprofile.columns)
    ind = goodprofile.index.tolist()
    corr = []
    for i, (k, v) in enumerate(repprofiles[a].iterrows()):
        h.showcount(i, len(repprofiles))
        res = np.array([np.corrcoef(v, w)[0, 1] for _, w in goodprofile[a].iterrows()])
        if max(res) == 1 or closest:
            match[k] = ind[np.argmax(res)]
        if returncorr:
            corr.append(res)
    if returncorr:
        corr = pd.DataFrame(
            data=corr,
            index=repprofiles.index.tolist(),
            columns=goodprofile.index.tolist(),
        )
        return match, corr
    else:
        return match


def loadGCTXasAnnData(path):
    val = path[:-5]
    rcmd = (
        'library("cmapR"); my_ds <- parse_gctx("'
        + val
        + '.gctx"); write.csv(my_ds@rdesc, "'
        + val
        + '_rdesc.csv"); write.csv(my_ds@cdesc, "'
        + val
        + '_cdesc.csv"); write.csv(my_ds@mat, file=gzfile("'
        + val
        + '_mat.csv.gz"))'
    )
    res = subprocess.run("R -e '" + rcmd + "'", shell=True, capture_output=True)
    print(res)
    rdesc = pd.read_csv(val + "_rdesc.csv", index_col=0).set_index("id")
    cdesc = pd.read_csv(val + "_cdesc.csv", index_col=0).set_index("id")
    # read as gz csv
    mat = pd.read_csv(val + "_mat.csv.gz", index_col=0, compression="gzip")
    res = AnnData(X=mat, var=cdesc, obs=rdesc)
    cmd = (
        "rm "
        + val
        + "_mat.csv.gz & rm "
        + val
        + "_rdesc.csv & rm "
        + val
        + "_cdesc.csv"
    )
    os.system(cmd)
    return res


def plot(
    tpm,
    ann,
    color_column="tissue_type",
    pca_comp=70,
    n_neighbors=10,
    min_dist=0.5,
    metric="euclidean",
    clusts=None,
    low_mem=False,
    colortable=None,
    **plot_kwargs,
):
    if len(tpm) > 50_000:
        print("lowmem mode")
        low_mem = True

    print("PCA...")
    tpm = (
        PCA(n_components=pca_comp).fit_transform(tpm)
        if not low_mem
        else IncrementalPCA(n_components=pca_comp)
    )
    print("UMAP...")
    tpm = umap.UMAP(
        n_neighbors=n_neighbors, min_dist=min_dist, metric=metric, n_components=2
    ).fit_transform(tpm)
    if ann is not None:
        # annotations to dict
        plot_kwargs["labels"] = {k: list(v) for k, v in ann.T.iterrows()}
    if clusts is not None:
        plot_kwargs["labels"].update({"clusters": clusts})
    if "colors" not in plot_kwargs:
        if clusts is not None:
            col = {l: i for i, l in enumerate(set(clusts))}
            plot_kwargs.update({"colors": [col[x] for x in clusts]})
        else:
            if colortable is None:
                col = {l: i for i, l in enumerate(set(ann[color_column]))}
            else:
                col = colortable
                plot_kwargs.update({"colprovided": True})
            plot_kwargs.update({"colors": [col[x] for x in ann[color_column].tolist()]})
    print("plot...")
    hplot.scatter(tpm, **plot_kwargs)



import numpy as np
from scipy.spatial import distance_matrix
from genepy.utils import helper as h
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd


# make a plot of averaged binned signal strength by distance from locis
def computeDistsFromClass(dots, seconddots, conds=['DMSO', 'VHL'], groupcol="group",
                          sclass='green', signal="mean_green", area="area"):
  """
  """
  dists= {}
  twodists = {}
  for val in set(dots.exp):
    for e in conds:
      d = dots[(dots.exp==val)&(dots.treat==e)]
      dist = []
      weight = []
      newdist = []
      ind=[]
      m = seconddots[(seconddots.exp==val)&(seconddots.treat==e)]
      print(val, e)
      for i,(k, v) in enumerate(m.iterrows()):
        h.showcount(i, len(m))
        dist.append(
          distance_matrix(d[(d['class']==sclass)&
                            (d[groupcol]==v[groupcol])][['x', "y", "z"]].values,
                          np.array([v[['x_mean', "y_mean", "z_mean"]]])).T[0].astype(float))
        weight.append(d[(d['class'] == sclass)&(d[groupcol]==v[groupcol])][signal])
        dat = d[(d['class'] == sclass) &
                                    (d[groupcol] == v[groupcol])][['x', "y", "z", signal, area, "m_id"]]
        a = dat.values
        a[:,:3] = a[:,:3] - v[['x_mean', "y_mean", "z_mean"]].values
        newdist.append(a)
        ind.extend(dat.index.tolist())
      twodists[val+e] = pd.DataFrame(data=np.vstack(newdist),
                                      columns=['x', 'y', 'z', signal, area, "m_id"],
                            index=ind)
      dists[val+e] = [np.hstack(dist), np.hstack(weight)]
  return twodists, dists


def drawDots(dists, scenter=False, size=1000, zsize=1000,
             folder="", signal="signal", levels=20,
             area="area", vmin=None, vmax=None,
              norm=None, norm_dots=None, second=None,
            color="seagreen",
            seccolor=sns.light_palette("orange", as_cmap=True), **kwargs):
  """
  """
  sm = []
  m = []
  sca=1.2
  if second is not None:
    for _, a in dists.items():
      sm.append(a[second(a)][signal].max())
  for _, a in dists.items():
    m.append(a[signal].mean())
  for i, (k,a) in enumerate(dists.items()):
    a = a.copy()
    a[area] = ((a[area]/(3.14))**(1/2)).astype(float)

    a = a[(abs(a.x)<size*sca) & (abs(a.y)<size*sca) & (abs(a.z)<zsize*sca)]
    #ax = sns.scatterplot(data=a, x='x', y='y', hue_norm=(None,max(m)) if norm is None else norm,
                      #	hue=signal, size=area, palette=color, **kwargs)
    if type(norm) is dict:
      n=norm[k]
    elif type(norm) is list:
      n=norm[i]
    else:
      n=None
    ax=sns.kdeplot(data=a[['x', 'y', signal]].astype(float),
                 x='x', fill=True, y='y', weights=signal, color=color,
                 thresh=False, levels=levels, cbar=m[i] == max(m),
                 hue_norm=n, vmin=vmin/(m[i]/max(m)), vmax=vmax/(m[i]/max(m)))
    # (None, max(sm)/sca) if norm is None else norm)
    if second is not None:
      print('adding second color')
      ex = sns.scatterplot(data=a[second(a)].sort_values(by=signal), x='x', y='y',
                        hue_norm=(None, max(sm)/sca) if norm_dots is None else norm_dots,
                        hue=signal, palette=seccolor, size=area, **kwargs)
      ex.legend(bbox_to_anchor=(2, 1), loc=1)
    plt.title(k)
    if scenter:
      ax.plot([0], [0], 'o', ms=scenter, markerfacecolor="None",
      markeredgecolor='red', markeredgewidth=1)

    ax.set_xlim((-size,size))
    ax.set_ylim((-size,size))
    plt.show()
    ax.get_figure().savefig(folder+k+"_scatter_representation_size_to_center.pdf")


def colocalize(dots, groupcol='group', zcol='z', xcol="x", ycol="y",
              default_maxdist=None, distance_scale=1.2, areacol="area",
              mergedidcol='m_id'):
  """
  """
  idcount = 0
  merged = dots.copy()
  for group in set(dots[groupcol]):
    print(group)
    # merging cells
    gdot = dots[(dots[groupcol] == group)]
    maxdist = default_maxdist if default_maxdist else np.sqrt(
        gdot[areacol].mean() / 3.14) * distance_scale

    gdot[mergedidcol]=None

    pos = gdot[[xcol, ycol, zcol]].values
    dist = distance_matrix(pos, pos)

    # closest needs to not be too far away otherwise the dot
    # is considered as finished
    for val in np.tril(dist < maxdist):
      con = np.argwhere(val > 0).T[0].tolist()
      # we get all its connections
      con_val = gdot.iloc[con]
      ids = list(set(con_val[mergedidcol]) - set([None]))
      # if connections are already connected we use this id
      if len(ids) > 0:
        def_id = ids[0]
        # for each connection, if have another id,
        # replace this id with the current one
        for i in ids:
          con.extend(np.argwhere(
              (gdot[mergedidcol] == i).values).T[0].tolist())
        con = list(set(con))
      # if none we create a new id
      else:
        idcount+=1
        def_id = "id_"+str(idcount)
      gdot.loc[gdot.iloc[con].index.tolist(), mergedidcol] = def_id
    #except:
    #	pdb.set_trace()
    merged.loc[gdot.index.tolist(), mergedidcol] = gdot[mergedidcol].tolist()
  return merged

def mergeAnnotated(annot, minzstack=2, groupdefault={}, todrop=[], coltocount="image",
                    id="m_id", colocName="cobinding"):
  """
  """
  annot = annot.drop(columns=todrop)
  grouping = {i: "mean" for i in annot.columns}
  if groupdefault:
    grouping.update(groupdefault)
  grouping.pop(id)
  # merge into a same sample
  groups = annot.groupby(id)
  counts = groups[coltocount].count()
  merged = groups.agg(grouping)
  merged['counts'] = counts
  merged = merged[merged['counts'] >= minzstack]
  merged.columns = [i[0] if "first" in i[1]
                  else '_'.join(i) for i in merged.columns]
  #rename colors
  merged['class'] = [i[0] if len(
    i) == 1 else colocName for i in merged["class_unique"]]
  return merged.drop(columns="class_unique")



import os
import pandas as pd
import numpy as np
from genepy.utils import helper as h
import re
import pyBigWig
from scipy.stats import zscore, fisher_exact
import subprocess
from statsmodels.stats.multitest import multipletests

size = {"GRCh37": 2864785220, "GRCh38": 2913022398}

cmaps = ['Greys', 'Purples', 'Blues', 'Greens', 'Oranges', 'Reds',
				'YlOrBr', 'YlOrRd', 'OrRd', 'PuRd', 'RdPu', 'BuPu',
				'GnBu', 'PuBu', 'YlGnBu', 'PuBuGn', 'BuGn', 'YlGn']

chroms = {'chr1', 'chr10', 'chr11', 'chr12', 'chr13', 'chr14', 'chr15', 'chr16', 'chr17', 'chr18', 'chr19',
'chr2', 'chr20', 'chr21', 'chr22', 'chr3', 'chr4', 'chr5', 'chr6', 'chr7', 'chr8', 'chr9', 'chrX',
 'chrY', '1', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '2', '20', '21', '22', '3', '4', '5', '6',
 '7', '8', '9', 'X', 'Y'}


def bigWigFrom(bams, folder="", numthreads=8, genome='GRCh37', scaling=None, verbose=1):
	"""
	run the bigwig command line for a set of bam files in a folder

	Can apply some scaling and process files in parallel

	Args:
	----
		bams: list[str] of filepath to bam files
		folder: str folder where the bam files would be stored (will save the bigwigs there as well, else to current path ./)
		numthreadds: int number of threads to process in parallel
		genome: str genome version to use (for genome size) only GRCh37 GRCh38 (see size global variable in this file)
		scaling: list[float] a list of scaling values to apply to the bigwigs. if provided, won't scale to read counts.
		verbose: int verbose level

	"""
	if "bigwig" not in os.listdir(folder if folder else '.'):
		os.mkdir(folder + "bigwig")
	for i, bam in enumerate(bams):
		in1 = folder + bam
		out1 = folder + "bigwig/" + bam.split('/')[-1].split('.')[0] + '.bw'
		cmd = "bamCoverage --effectiveGenomeSize " + str(size[genome]) + " -p " + str(numthreads) +\
			" -b " + in1 + " -o " + out1
		if scaling is not None:
			cmd += ' --scaleFactor ' + str(scaling[i])
		if verbose == 0:
			cmd += ' 2> ' + bam + '.error.log'
		res = subprocess.run(cmd, capture_output=True, shell=True)
		if res.returncode != 0:
			raise ValueError('issue with the command: ' + str(res.stderr))


def ReadRoseSuperEnhancers(roseFolder, containsINPUT=True, name="MV411"):
	"""
	reads ROSE2's output and returns its superenhancer bedfile as a pd dataframe. 

	Can be multiple superenhancers from a set of HRK27ac bam files

	Args:
	-----
		roseFolder: str folderpath to ROSE's output
		containsINPUT: bool whether the bedfile contains INPUT signal as well
		name: str sample name from which we computed superenhancers

	Returns:
	--------
		a dataframe of the bed representation of the superenhancers
	"""
	beds = os.listdir(roseFolder)
	superenhan = pd.DataFrame()
	for i in beds:
		if i.endswith('_SuperEnhancers.table.txt'):
			superenhan = superenhan.append(pd.read_csv(roseFolder+i, sep='\t', skiprows=5)).drop(columns=['enhancerRank','isSuper'])
	data = [superenhan.columns[6]] + superenhan.columns[8:].tolist()
	if containsINPUT:
		inputd = superenhan.columns[7]
	superenhan['name'] = [i.split(name)[1].split('_')[2] for i in superenhan['REGION_ID']]
	superenhan = superenhan.rename(columns={'CONSTITUENT_SIZE':'size','REGION_ID':"id",'CHROM':'chrom','START':'start','STOP':'end',"NUM_LOCI":'num'}).replace(np.nan,0)
	superenhan['foldchange'] = superenhan[data].sum(1)/superenhan[inputd]
	superenhan = superenhan.drop(columns=data+[inputd])
	return superenhan.sort_values(by=['chrom','start','end'])


def loadPeaks(peakFile=None, peakfolder=None, isNF=True, CTFlist=[], skiprows=0):
	"""
	loads 1 to many peak bedfile into one pandas dataframe.

	all og the peaks will be concatenated into one dataframe. this function can 
	work with jkobject|nfcore/chipseq nextflow pipelines output (isNF)

	Args:
	-----
		peakFile: str filepath to a peak bedfile
		peakfolder: str folderpath to a folder congaining beddfile
		isNF: bool if the peaks come from an nfcore/chipseq nextflow pipeline (with VA_names)
		CTFlist: list[str] only load peaks corresponding from this list of names
		skiprows: int if there is rows to skip in the file (can be 1, often)

	Returns:
	-------
		a bed like dataframe containing the concatenated peaks of the bed files
	"""
	if peakfolder:
		bindings = pd.DataFrame()
		for folder in os.listdir(peakfolder):
			if isNF:
				if any(tf in folder for tf in CTFlist) or not CTFlist:
					binding = pd.read_csv(peakfolder + folder + "/NA_peaks.narrowPeak", sep='\t', header=None) if\
					os.exists(peakfolder + folder + "/NA_peaks.narrowPeak") else peakfolder + folder + "/NA_peaks.broadPeak"
					binding['name'] = folder.replace('.narrowPeak', '').replace('.broadPeak','')
					bindings = bindings.append(binding)
			else:
				file = folder
				if file[-10:] in ["narrowPeak",".broadPeak"] and (any(tf in file for tf in CTFlist) or not CTFlist):
					print('reading: '+file)
					binding = pd.read_csv(peakfolder + file, sep='\t', header=None, skiprows=skiprows)
					binding['name'] = file.replace('.narrowPeak', '').replace('.broadPeak','')
					bindings = bindings.append(binding)
	elif peakFile:
		bindings = pd.read_csv(peakFile, sep='\t', header=None, skiprows=skiprows)
		bindings['name'] = peakFile.split('/')[-1].split('.')[0]
	else:
		raise ValueError("need to provide one of peakFile or peakfolder")
	if not len(bindings.columns)==6:
		bindings = bindings.drop(5, 1).drop(4, 1).rename(columns={6:4, 7:5, 8:6, 9:7,})
	bindings = bindings.rename(columns={
		0: "chrom",
		1: 'start',
		2: 'end',
		3: 'peak_number',
		4: "foldchange",
		5: "-log10pvalue",
		6: "-log10qvalue",
		7: 'relative_summit_pos'})
	bindings = bindings.sort_values(by=["chrom", "start", "end"], axis=0)
	bindings.start = bindings.start.astype('int')
	bindings.end = bindings.end.astype('int')
	if not len(bindings.columns)==6:
		bindings['relative_summit_pos'] = bindings.relative_summit_pos.astype(
			float) if 'relative_summit_pos' in bindings.columns else bindings.end - bindings.start
		bindings["-log10pvalue"] = bindings["-log10pvalue"].astype('float')
		bindings['-log10qvalue'] = bindings['-log10qvalue'].astype('float')
		loc = bindings['relative_summit_pos'].isna()
		bindings.loc[bindings[loc].index, 'relative_summit_pos'] = bindings[loc].end - bindings[loc].start
		bindings.relative_summit_pos = bindings.relative_summit_pos.astype(int)
	bindings.foldchange = bindings.foldchange.astype('float')
	bindings = bindings.reset_index(drop=True)
	return bindings


def simpleMergePeaks(peaks, window=0, totpeaknumber=0, maxp=True, mergedFold="mean"):
	"""
	simply merges bedfiles from peak callers. providing a concaneted dataframe of bed-like tables

	will recompute pvalues and foldchange from that.

	Args:
	-----
		peaks: pd.df of concatenated bed-like dataframe with a name column differentiating each peak group
			to merge
		window: int the max size between each peaks under which to still merge the peaks together 
		totpeaknumber: int, set to more than 0 for peaks to be counted from this value 
			(e.g. 10, with 100 peaks will give peak_number of 10-110)
		maxp: bool, set to true to merge pvalue uusing the max pvalue, else will take the product of all
		mergedFold: flag "mean"|"max"|"sum", on how to merge foldchanges

	Returns:
	-------
		df bed-like of the merged peaks
	"""
	peaks = peaks.sort_values(by=['chrom', 'start','end'])
	tfs = list(set(peaks['name']))
	mergedpeaksdict = {}
	remove = []
	peaknumber = 0
	merged_bed = {
		"chrom": [peaks.iloc[0]['chrom']],
		"start": [peaks.iloc[0]['start']],
		"end": [],
		"peak_number": [peaknumber + totpeaknumber],
		"foldchange": [],
		"-log10pvalue": [],
		"-log10qvalue": [],
		"relative_summit_pos": []
	}
	foldchange = [peaks.iloc[0].get('foldchange', 1)]
	log10pvalue = [peaks.iloc[0].get('-log10pvalue', 0)]
	log10qvalue = [peaks.iloc[0].get('-log10qvalue', 0)]
	relative_summit_pos = peaks.iloc[1].get('relative_summit_pos', peaks.iloc[0]['start'])
	# computes overlap by extending a bit the window (100bp?) should be ~readsize
	prev_end = peaks.iloc[0]['end']
	prev_chrom = peaks.iloc[0]['chrom']
	tfmerged = {a: [0] for a in tfs}
	tfmerged[peaks.iloc[0]['name']][-1] = peaks.iloc[0].get('foldchange', 1)
	for i, (pos, peak) in enumerate(peaks.iloc[1:].iterrows()):
		print(str(i / len(peaks)), end="\r")
		if prev_end + window > peak['start'] and prev_chrom == peak['chrom']:
			# can be merged
			if peak.get('foldchange', 1) > max(foldchange):
				relative_summit_pos = peak.get('relative_summit_pos', peaks['start'])
			foldchange.append(peak.get('foldchange', 1))
			log10pvalue.append(peak.get('-log10pvalue', 0))
			log10qvalue.append(peak.get('-log10qvalue', 0))

		else:
			# newpeak
			for k, val in tfmerged.items():
				val.append(0)
			peaknumber += 1
			merged_bed['chrom'].append(peak['chrom'])
			merged_bed['start'].append(peak['start'])
			merged_bed['end'].append(prev_end)
			merged_bed['peak_number'].append(peaknumber + totpeaknumber)
			if mergedFold=="mean":
				merged_bed['foldchange'].append(np.mean(foldchange))
			elif mergedFold=="max":
				merged_bed['foldchange'].append(max(foldchange))
			elif mergedFold=="sum":
				merged_bed['foldchange'].append(sum(foldchange))
			else:
				raise ValueError("mergedFold needs to be one of:")
			merged_bed['-log10pvalue'].append(max(log10pvalue) if maxp else np.prod(log10pvalue))
			merged_bed['-log10qvalue'].append(max(log10qvalue) if maxp else np.prod(log10qvalue))
			merged_bed['relative_summit_pos'].append(relative_summit_pos)
			foldchange = [peak.get('foldchange', 1)]
			log10pvalue = [peak.get('-log10pvalue', 0)]
			log10qvalue = [peak.get('-log10qvalue', 0)]
			relative_summit_pos = peak.get('relative_summit_pos', peak['start'])
		prev_end = peak['end']
		prev_chrom = peak['chrom']
		tfmerged[peak['name']][-1] = peak.get('foldchange', 1)
	merged_bed['end'].append(prev_end)
	if mergedFold=="mean":
		merged_bed['foldchange'].append(np.mean(foldchange))
	elif mergedFold=="max":
		merged_bed['foldchange'].append(max(foldchange))
	elif mergedFold=="sum":
		merged_bed['foldchange'].append(sum(foldchange))
	else:
		raise ValueError("mergedFold needs to be one of:")
	merged_bed['-log10pvalue'].append(max(log10pvalue) if maxp else np.prod(log10pvalue))
	merged_bed['-log10qvalue'].append(max(log10qvalue) if maxp else np.prod(log10qvalue))
	merged_bed['relative_summit_pos'].append(relative_summit_pos)

	merged_bed = pd.DataFrame(merged_bed)
	tfmerged = pd.DataFrame(tfmerged)
	return pd.concat([merged_bed, tfmerged], axis=1, sort=False)

#from numba import jit, float32, int8

#@jit(float32[:](float32[:, 3], float32[:, 4], int8, str))
def putInBed(conscensus, value, window=10, mergetype='mean', columns=['foldchange']):
	""" 
	given a conscensus bed-like dataframe and another one, will merge the second one into the first

	Args:
	-----
		conscensus df[start,end,chrom] the conscensus (first one)
		value df[start, end, chrom]+columns the value one (second one)
		mergetype: flag: mean,first,last, on how to merge ttwo peaks that would fall on the same one
		window: int on max distance of second df peaks' to the first ones, to still merge them
			re.g. 0 is hard overap, 10000000 is nearest peak)

	Returns:
	-------
		np.array of values of the second dataframe to add to the first one 
		(e.g. can do df1[newcol] = returned_array)
	"""
	conscensus = conscensus.sort_values(by=['chrom','start','end']).reset_index(drop=True)
	value = value.sort_values(by=['chrom','start','end']).reset_index(drop=True)
	locinvalue=0
	loc=0
	tot=0
	num = []
	res = np.zeros((len(conscensus), len(columns)))
	not_end=True
	def add(res,num,not_end,loc):
		if len(num)>0:
			if mergetype=='mean':
				res[loc,:] = np.mean(num, 0)
			elif mergetype=='sum':
				res[loc,:]=np.sum(num, 0)
			elif mergetype=='first':
				res[loc,:]=num[0]
			elif mergetype=='last':
				res[loc,:]=num[-1]
			else:
				raise ValueError('must be one of')
			num= []
		loc+=1
		if loc == len(conscensus):
			not_end=False
		return res, num, not_end,loc
	while not_end:
		print(loc/len(conscensus),end="\r")
		try:
			a = conscensus.iloc[loc]
			b = value.iloc[locinvalue]
			# we need to switch chromosome for b
			if b.chrom < a.chrom:
				locinvalue+=1
				if locinvalue == len(value):
					not_end=False
			# we need to switch chromosome for b
			elif b.chrom > a.chrom:
				loc+=1
				if loc == len(conscensus):
					not_end=False
			elif b.start<a.start:
				if b.end+window>a.start:
					tot+=1
					num.append(b[columns])
					if b.end>a.end+window:
						res,num,not_end,loc = add(res,num,not_end,loc)
						continue
				locinvalue+=1
				if locinvalue == len(value):
					not_end=False
			elif b.start<a.end+window:
				tot+=1
				num.append(b[columns])
				if b.end>a.end+window:
					res,num,not_end,loc = add(res,num,not_end,loc)
					continue
				locinvalue+=1
				if locinvalue == len(value):
					not_end=False
			else:
				res,num,not_end,loc = add(res,num,not_end,loc)
		except:
			import pdb; pdb.set_trace()
	print(str(tot)+' were merged into conscensus')
	return res


async def pairwiseOverlap(bedfile, norm=True, bedcol=8, correct=True, docorrelation=True):
	"""
	compute pairwise overlap and correlation on this overlap for a set of peaks mappe to a conscensus 

	with each columns after the 7th one representing the signal of a given ChIP experiment
	over this conscensus. will present overlap of row values in col values

	Args:
	----
		bedfile: df bed-like representing a conscensus set of peaks, and a set of values/foldchanges
			over it
		norm: bool wether to zscore normalise the signal or not
		bedcol: int col where the bed information ends and signal information columns start
		correct: bool whether to correct for fully similar lines/ columns by removing 1 on their last value
		docorrelation: bool whether or not to compute correlation as well as enrichment

	Returns:
	-------
		a dataframe[values_name x values_name] of % overlap of row values in col values in the consensus
		a dataframe[values_name x values_name] of correlation of values signal over the overlaps
	"""
	if correct:
		print("we will be correcting for fully similar lines/ columns by removing 1 on their last value")
	dat = bedfile[bedfile.columns[bedcol:]].values
	correlation = np.ones((dat.shape[1],dat.shape[1]))
	overlap = np.ones((dat.shape[1],dat.shape[1]))
	for i, col in enumerate(dat.T):
		#pdb.set_trace()
		overlapping = np.delete(dat,i,axis=1)[col!=0]
		col = col[col!=0]
		add=0
		for j, val in enumerate(overlapping.T):
			if j==i:
				add=1
			if docorrelation:
				if norm and not np.isnan(zscore(col[val!=0]).sum()) and not np.isnan(zscore(val[val!=0]).sum()) or not correct:
					correlation[i,j+add] = np.corrcoef(zscore(val[val!=0]),zscore(col)[val!=0])[0,1]
				else:
					tmp = np.corrcoef(val[val != 0], col[val != 0])[0, 1]
					if np.isnan(tmp) and correct:
						if len(col[val!=0]) == 0 or len(val[val!=0]) == 0:
						# one has no overlap
							correlation[i,j+add] =  0
						else:
							# one contains only the same value everywhere
							col[-1]-=max(0.01,abs(np.mean(col)))
							val[-1]-=max(0.01,abs(np.mean(val)))
							correlation[i,j+add] = np.corrcoef(val,col)[0,1]
			overlap[i,j+add]=len(val[val!=0])/len(col)
	if docorrelation:
		correlation = pd.DataFrame(data=correlation, index=bedfile.columns[bedcol:], columns=bedfile.columns[bedcol:])
		correlation[correlation.isna()] = 0
	overlap = pd.DataFrame(data=overlap, index=bedfile.columns[bedcol:], columns=bedfile.columns[bedcol:]).T
	return overlap, correlation if docorrelation else None


async def enrichment(bedfile: pd.DataFrame, bedcol=8, groups=None, correct=True, okpval=10**-3):
	"""
	compute pairwise enrichment and correlation for a set of peaks mappe to a conscensus 

	with each columns after the 7th one representing the signal of a given ChIP experiment
	over this conscensus. will present enrichment of row values in col values

	Args:
	----
		bedfile: df bed-like representing a conscensus set of peaks, and a set of values/foldchanges
			over it
		bedcol: int col where the bed information ends and signal information columns start
		correct: bool whether to correct for multiple hypothesis testing or not
		docorrelation: bool whether or not to compute correlation as well as enrichment
		okpval: float max pvalue over which to set the enrichment to 0

	Returns:
	-------
		a dataframe[values_name x values_name] of enrichment of row values in col values
		a dataframe[values_name x values_name] of correlation of values signal over the overlaps
	"""
	dat = bedfile[bedfile.columns[bedcol:]].values
	prob = dat.astype(bool).sum(0)/len(dat)
	enrichment = np.zeros((dat.shape[1] if groups is None else len(set(groups)), dat.shape[1]))
	pvals = np.zeros(
		(dat.shape[1] if groups is None else len(set(groups)), dat.shape[1]))
	if groups is not None:
		for i in set(groups):
			overlapping = dat[groups==i]
			for j,val in enumerate(overlapping.T):
				# enrichment of j in i
				e, p = fisher_exact([
					[len(val[val != 0]), len(val[val == 0])],
					[prob[j]*len(dat), (1-prob[j])*len(dat)]])
				enrichment[i, j] = np.log2(e)
				pvals[i, j] = p
	else:
		for i, col in enumerate(dat.T):
			overlapping = np.delete(dat,i,axis=1)[col!=0]
			col = col[col!=0]
			add=0
			for j, val in enumerate(overlapping.T):
				if j==i:
					add=1
					enrichment[i,i]=0
				e, p = fisher_exact([[len(val[val != 0]), len(val[val == 0])], [
					prob[j+add]*len(dat), (1-prob[j+add])*len(dat)]])
				enrichment[i, j+add] = np.log2(e)
				pvals[i, j+add] = p
		enrichment[i,i]=0
	enrichment = pd.DataFrame(data=enrichment, index=bedfile.columns[bedcol:] if groups is None else set(groups), columns=bedfile.columns[bedcol:]).T
	enrichment[enrichment==-np.inf] = -1000
	enrichment[enrichment.isna()] = 0
	enrichment[enrichment == np.inf] = 1000
	if correct:
		pvals = np.reshape(multipletests(pvals.ravel(),
										0.1, method="bonferroni")[1], pvals.shape)
	pvals = pd.DataFrame(
		data=pvals, index=bedfile.columns[bedcol:]  if groups is None else set(groups), columns=bedfile.columns[bedcol:]).T
	enrichment[pvals>okpval] = 0
	return enrichment, pvals


def findAdditionalCobindingSignal(conscensus, known=None, bigwigs=[], window=100):
	"""
	somewhat similar concept to computePeaksAt

	# get pysam data
	# ask for counts only at specific locus based on peak windows from mergedpeakset
	# append to an array
	# return array, normalized
	"""
	raise NotImplementedError()
	if known:
		print('getting '+ str(len(peaks.tf))+' peaks. Using the peaks values directly if \
				available and using the bigwigs otherwise.')
		res = known.values.astype(float)
	elif len(bigwigs)>0:
		print('getting '+str(len(bigwigs))+' bigwigs, no peaks passed. Will compute the cobinding values\
			across the conscensus for each bigwigs.')
		res = np.zeros((len(bigwigs), len(conscensus)), dtype=float)
	else:
		raise ValueError('you need to pass a list of path to bigwigs for each/some samples')
	for i, bw in enumerate(bigwigs):
		if known:
			found = False
			for j, val in enumerate(known.tf):
				if val in bw:
					if found:
						raise ValueError('found two or more matching tf for bigwig: '+str(bw))
					found = True
					i=j
					break
				if not found:
					print('no tf found in known for tf: '+bw)
					raise ValueError('you need to have an amount of known columns equal to your bigwigs')
		print('doing file ' + str(bw))
		bw = pyBigWig.open(bw)
		for k, val in conscensus.iterrows():
			if known:
				if res[i][k]!=0:
					continue
			start = max([val.start - window, 0])
			end = min(val.end + window, bw.chroms(str(val.chrom)))
			res[i][k] = bw.stats(str(val.chrom), start, end)[0]
	res = np.nan_to_num(res, 0)
	return conscensus.join(pd.Dataframe(data=(res.T / res.max(1)).T, columns=bigwigs if not known else known.columns))



def annotatePeaks():
	"""
	get H3k27AC peaks and compute Rose
	for each TF peak
	assign super enhancer if within it and create a super enhancer TFgroup
	for each peaks
	apply similar to rose and merge TF together. (create new TFgroup)
	for each TF groups say
		if within super enhancer
		if within high h3k27ac marks
		its nearest gene (distance to it, cis/trans)


	FIND A WAY TO FILTER TO ONLY PLACES WITH H3K27AC marks??

	TAD points where most contacts on one side happened on this side.
	specific distance. zone of most concentration of contacts for a peak region.
	"""
	# if way == 'Closest':

	# elif way == 'ClosestExpressed':

	# elif way == 'ActivityByContact':

	# else:
	#     raise ValueError('needs to be oneof Closest ClosestExpressed ActivityByContact')
	raise NotImplementedError()


# os.system()


def getCoLocalization():
	"""
	for each annotations (super enhancer & TFgroups)
	for each TF, find highest peak/meanCov. if above thresh add to localization
	"""
	raise NotImplementedError()


def refineGroupsWithHiC():
	"""
	given HiC data, for each loops (should be less than X Mb distance. create a Xkb zone around both sides
	and find + merge each TF/TFgroup at this location)
	"""
	raise NotImplementedError()


async def fullDiffPeak(bam1, bam2, control1, size=None, control2=None, scaling=None, directory='diffData/',
				 res_directory="diffPeaks/", isTF=False, compute_size=True, pairedend=True):
	"""
	will use macs3 to call differential peak binding from two bam files and their control

	one can also provide some spike in scaling information

	Args:
	-----
	bam1: str bamfilepath
	bam2: str bamfilepath
	control1: str control bam filepath (INPUT/IGG)
	control2: str control bam filepath (INPUT/IGG)
	scaling: tuple(float1,float2) of scaling of each bams 
	size: int the read length (extsize parameter)
	directory: str the directory where to save the produced data
	res_directory: str the directory where to save the results
	isTF: bool true if TF else false (will change the size)
	compute_size: bool whether to compute the extsize value with `macs3 predictd`
	pairedend: bool if paired end

	Returns:
	-------
		str the log of the macs3 bdgdiff command
	"""
	print("doing diff from " + bam1 + " and " + bam2)
	if scaling is not None:
		if max(scaling) > 1:
			raise ValueError("scalings need to be between 0-1")
	name1 = bam1.split('/')[-1].split('.')[0]
	name2 = bam2.split('/')[-1].split('.')[0]
	if size is None:
		if isTF:
			size = 147
		else:
			size = 200
	if compute_size:
		print('computing the fragment avg size')
		cmd = "macs3 predictd -i " + bam1
		ret = subprocess.run(cmd, capture_output=True, shell=True)
		size = re.findall("# predicted fragment length is (\d+)", str(ret.stderr))[0]
		print(size)
	else:
		print('using default|given size')
	pairedend = "BAMPE" if pairedend else "BAM"
	if control2 is None:
		control2 = control1
	cmd1 = "macs3 callpeak -B -t " + bam1 + " -c " + control1 + " --nomodel --extsize " + str(size) + " -n " + name1 + " --outdir " + directory + " -f " + pairedend
	cmd2 = "macs3 callpeak -B -t " + bam2 + " -c " + control2 + " --nomodel --extsize " + str(size) + " -n " + name2 + " --outdir " + directory + " -f " + pairedend
	print('computing the scaling values')
	ret = subprocess.run(cmd1, capture_output=True, shell=True)
	print(ret.stderr)
	scaling1a = int(re.findall(" after filtering in treatment: (\d+)", str(ret.stderr))[0])
	scaling1b = int(re.findall(" after filtering in control: (\d+)", str(ret.stderr))[0])
	scaling1 = scaling1a if scaling1a <= scaling1b else scaling1b
	ret = subprocess.run(cmd2, capture_output=True, shell=True)
	print(ret.stderr)
	scaling2a = int(re.findall(" after filtering in treatment: (\d+)", str(ret.stderr))[0])
	scaling2b = int(re.findall(" after filtering in control: (\d+)", str(ret.stderr))[0])
	scaling2 = scaling2a if scaling2a <= scaling2b else scaling2b
	if scaling is not None:
		scaling1 = int(scaling1/scaling[0])
		scaling2 = int(scaling2/scaling[1])
	print(scaling1, scaling2)
	return (await diffPeak(directory+name1+"_treat_pileup.bdg", directory+name2+"_treat_pileup.bdg",
		directory+name1+"_control_lambda.bdg", directory+name2+"_control_lambda.bdg",
		res_directory, scaling1, scaling2, size))


async def diffPeak(name1, name2, control1, control2, res_directory, scaling1, scaling2, size):
	"""
	calls MACS2 bdgdiff given some parameters

	Args:
	-----
		name1: str bamfilepath
		name2: str bamfilepath
		control1: str control bam filepath (INPUT/IGG)
		control2: str control bam filepath (INPUT/IGG)
		res_directory: str the directory where to save the results
		scaling1: float the scaling factor of the first bam
		scaling2: float the scaling factor of the second bam
		size: int the read length (extsize parameter)

	Returns:
	-------
		str the log of the macs3 bdgdiff command
	"""
	print("doing differential peak binding")
	cmd = "macs3 bdgdiff --t1 " + name1 + " --c1 "
	cmd += control1+" --t2 " + name2 +" --c2 " + control2
	cmd += " --d1 " + str(scaling1) + " --d2 " + str(scaling2) + " -g 60 "
	cmd += "-l " + str(size) + " --o-prefix " + name1.split('/')[-1].split('.')[0] + "_vs_"
	cmd += name2.split('/')[-1].split('.')[0] + " --outdir " + res_directory
	res = subprocess.run(cmd, capture_output=True, shell=True)
	return res


def AssignToClosestExpressed(bed,countFile,genelocFile):
	print("the bed file and genelocFile should use the same assembly")
	genelocFile = pd.read_csv(genelocFile,sep="\t", compression="", columns=['chrom','start','end',"name"])
	#for val in bed.iterrows():
	raise NotImplementedError()



async def makeSuperEnhancers(MACS2bed, bamFile, outdir, baiFile=None, rosePath=".",
	stitching_distance=None, TSS_EXCLUSION_ZONE_SIZE="2500", assembly="hg38",controlBam=None,controlBai=None):
	"""
	Calls super enhancer from H3K27ac with the ROSE algorithm

	Args:
	----
		MACS2bed: str he bed filepath to a MACS2 called set of peaks for an H3K27ac bam file
		bamFile: str bam filepath of the H3K27ac bam file
		outdir: str folderpath of the output. has to be an absolute path or a ~/path
		baiFile: str bam filepath of the H3K27ac bam file
		rosePath: str path to the ROSE folder containing the rose algorithm
		stitching_distance: int max stitching distance
		TSS_EXCLUSION_ZONE_SIZE: int size around TSS to excllude from SEs
		assembly: flag one of hg38, hg37...
		controlBam: str filepath to input bam file
		controlBai: str filepath to input bai file

	Returns:
	--------
		a bed-like dataframe of the superenhancers

	"""
	print("we are going to move your input files to "+rosePath)
	cmd = 'cp '+MACS2bed+' '+rosePath+MACS2bed.split('/')[-1]+'.bed'
	cmd += ' && mv '+bamFile+' '+rosePath
	baiFile = baiFile if baiFile else bamFile[:-1]+'i'
	cmd += ' && cp '+baiFile +' '+rosePath
	res = subprocess.run(cmd, capture_output=True, shell=True)
	if res.returncode != 0:
		raise SystemError('failed moving files: '+str(res))
	cmd = "cd "+rosePath+" && python ROSE_main.py -g "+assembly+" -i "+MACS2bed.split('/')[-1]+'.bed' + " -r " + bamFile.split('/')[-1] + " -o " + outdir
	if TSS_EXCLUSION_ZONE_SIZE:
		cmd+=" -t "+TSS_EXCLUSION_ZONE_SIZE
	if controlBam:
		os.system('mv '+controlBam+' '+rosePath)
		controlBai = controlBai if controlBai else controlBam[:-1]+'i'
		os.system('cp '+controlBai+' '+rosePath)
		cmd+=" -c "+controlBam.split('/')[-1]
	if stitching_distance:
		cmd+=" -s "+ stitching_distance

	res = subprocess.run(cmd, capture_output=True, shell=True)
	fail = False
	if res.returncode != 0:
		v = 'ROSE failed:' +str(res)
		fail = True
	print('finished.. moving them back to their original folder')
	cmd = 'rm '+rosePath+MACS2bed.split('/')[-1]+'.bed'
	cmd += ' && mv '+rosePath+bamFile.split('/')[-1]+' '+bamFile
	cmd+= ' && rm '+rosePath+baiFile.split('/')[-1]
	if controlBam:
		cmd += ' && mv '+rosePath+controlBam.split('/')[-1]+' '+controlBam
		cmd += ' && rm '+rosePath+controlBai.split('/')[-1]
	res = subprocess.run(cmd, capture_output=True, shell=True)
	if res.returncode != 0:
		raise SystemError('failed moving files: '+str(res))
	if fail:
		raise SystemError(v)
	print('worked')
	print(res)
	return ReadRoseSuperEnhancers(outdir ,bool(controlBam))



def runChromHMM(outdir, data, numstates=15, datatype='bed', folderPath=".", 
chromHMMFolderpath="~/ChromHMM/", assembly="hg38",control_bam_dir=None):
	"""
	runs the chromHMM algorithm

	Args:
	-----
		outdir str: an existing dir where the results should be saved
		data: df[cellname,markname,markbed|bam|bigwig, ?controlbed|bam|bigwig]
		numstates: number of states to use
		datatype: flag one of bed
		folderPath: str folder where to save chromHMM's work
		chromHMMFolderpath: str folderpath to chromHMM algorithm
		assembly: flag one of hg38, hg37 ...
		control_bam_dir: str directory where the control would be stored (if not given in the ddf)

	Returns:
	-------
		A dict of bed like dataframes containing the regions of the different states
	"""
	print("you need to have ChromHMM")
	chromHMM = "java -mx8000M -jar "+chromHMMFolderpath+"ChromHMM.jar "
	h.createFoldersFor(outdir+'binarized/')
	data.to_csv(outdir+"input_data.tsv", sep='\t',index=None,header=None)
	cmd = chromHMM
	if datatype=="bed":
		cmd+="BinarizeBed "
	elif datatype=="bigwig":
		cmd+="BinarizeSignal "
	elif datatype=="bam":
		cmd+="BinarizeBam "
	else:
		raise ValueError('you need to provide one of bam, bigwig, bed')
	cmd+= chromHMMFolderpath+"CHROMSIZES/"+assembly+".txt "+ folderPath+" "+outdir+"input_data.tsv "+outdir+"binarized"
	if control_bam_dir:
		cmd+=" -c "+control_bam_dir
	res1 = subprocess.run(cmd, capture_output=True, shell=True)
	print(res1)
	if res1.returncode!=0:
		raise ValueError(str(res1.stderr))
	cmd = chromHMM + "LearnModel -printposterior -noautoopen "
	if len(data)<10:
		cmd += '-init load -m '+chromHMMFolderpath+'model_15_coreMarks.txt '
	cmd += outdir+"binarized "+outdir+" "+str(numstates)+" "+assembly
	res2 = subprocess.run(cmd, capture_output=True, shell=True)
	print(res2)
	if res2.returncode!=0:
		raise ValueError(res2.stderr)
	ret = {}
	for v in set(data[0]):
		ret[v] = pd.read_csv(outdir+v+'_'+str(numstates)+'_dense.bed', sep='\t', header=None,
			skiprows=1).drop(columns=[4,5,6,7]).rename(columns=
			{0:'chrom',1:'start',2:'end',3:'state',8:"color"})
	return ret


def loadMEMEmotifs(file, tfsubset=[],motifspecies='HUMAN'):
	"""
	loads motif from the output file of MEME after running fimo.

	Args:
	----
		file: str location of the fimo gff|bed file
		tfsubset: list[str] subset of tf to load
		motifspecies: str name of the specied concerned (should be the same one as provided to MEME)

	Returns:
	--------
		df bed-like of each motifs across the genome
	"""
	if file.endswith('.gff'):
		print('converting to bed, you need to have "gfftobed" installed')
		cmd = 'gff2bed < '+file+' > '+file+'.bed'
		file = file+'.bed'
		res = subprocess.run(cmd,capture_output=True, shell=True)
		if res.returncode != 0:
			raise ValueError('issue with the command: ' + str(res.stderr))
		else:
			print(res.stdout.decode("utf-8"))
	## What are the motifs of our CRC members in ATACseq but not in our matrix
	merged_motif = pd.read_csv(file, sep='\t',skiprows=0,index_col=None, names=['pos',"fimo",
		"nucleotide_motif","relStart","relEnd","pval","strand",".","data"])
	merged_motif['tf']=[i[5:].split("_"+motifspecies)[0] for i in merged_motif.data]
	if tfsubset:
		merged_motif = merged_motif[merged_motif.tf.isin(tfsubset)]
	merged_motif['chrom'] = [i.split(':')[0][3:] for i in merged_motif.index]
	merged_motif['start'] = [i.split(':')[1].split('-')[0] for i in merged_motif.index]
	merged_motif['end'] = [i.split(':')[1].split('-')[1] for i in merged_motif.index]
	merged_motif = merged_motif.reset_index(drop=True)
	merged_motif = merged_motif.rename(columns={'pos':'relStart','fimo':'relEnd','nucleotide_motif':'pos',
		'relStart':'pval','relEnd':'strand','pval':'fimo','strand':'motif'})
	merged_motif['motif'] = [i.split('sequence=')[1].split(';')[0] for i in merged_motif.data]
	merged_motif['p_val'] = [i.split('pvalue=')[1].split(';')[0] for i in merged_motif.data]
	merged_motif['q_val'] = [i.split('qvalue=')[1].split(';')[0] for i in merged_motif.data]
	merged_motif = merged_motif.drop(columns=['pos','.','fimo','data'])
	merged_motif = merged_motif[merged_motif.columns[[6,7,8,0,1,3,2,9,10,5,4]]]
	merged_motif = merged_motif.sort_values(by=['chrom','start','end']).reset_index(drop=True)
	return merged_motif


def simpleMergeMotifs(motifs, window=0):
	"""
	aggregates the motifs if they overlap, into one motif file

	Args:
	----
		motifs: df bed-like of motifs locations
		window: int maxsize around motif for which to still merge

	Returns:
	--------
		df bedlike of merged motif:
		df bedlike of motifs that were not merged as they were different but still overlapping
	"""
	if type(motifs) is list:
		motifs = pd.concat(motifs)
	motifs = motifs.sort_values(by=['chrom', 'start'])
	toremove = []
	issues = []
	prevmotif = motifs.iloc[0]
	for i, (pos, motif) in enumerate(motifs.iloc[1:].iterrows()):
		print(str(i / len(motifs)), end="\r")
		if prevmotif['end'] + window > motif['start'] and prevmotif['chrom'] == motif['chrom']:
			# can be merged
			if motif['tf']!= prevmotif['tf'] or motif['motif'] != prevmotif['motif']:
				print('found different motifs overlapping')
				issues.extend([motif,prevmotif])
			else:
				toremove.append(pos)
		prevmotif = motif
	motifs = motifs.drop(index=toremove).reset_index(drop=True)
	issues = pd.concat(issues)
	return motifs, issues


def substractPeaksTo(peaks,loci, bp=50):
	"""
	removes all peaks that are not within a bp distance to a set of loci

	Args:
	----
		peaks: a bed file df with a chrom,start, end column at least
		loci: a df witth a chrom & loci column
		bp: the max allowed distance to the loci

	Returns:
	-------
		all the peaks that are within this distance
	"""
	i=0
	j=0
	keep=[]
	bp=50
	while j<len(peaks) and i<len(loci):
		h.showcount(j,len(peaks))
		if peaks.loc[j].chrom > loci.loc[i].chrom:
			i+=1
			continue
		if peaks.loc[j].chrom < loci.loc[i].chrom:
			j+=1
			continue
		if peaks.loc[j].start - bp > loci.loc[i].loci:
			i+=1
			continue
		if peaks.loc[j].end + bp< loci.loc[i].loci:
			j+=1
			continue
		if peaks.loc[j].end + bp >= loci.loc[i].loci and peaks.loc[j].start - bp <= loci.loc[i].loci:
			keep.append(j)
			j+=1
	return peaks.loc[set(keep)]

	



# Jeremie Kalfon
# for BroadInsitute
# in 2019

from __future__ import print_function
from multiprocessing.sharedctypes import Value
import os
import signal
import re

import pandas as pd
import numpy as np

from genepy.google import gcp
from genepy.utils import helper as h
from tqdm import tqdm

size = {"GRCh37": 2864785220, "GRCh38": 2913022398}

cmaps = [
    "Greys",
    "Purples",
    "Blues",
    "Greens",
    "Oranges",
    "Reds",
    "YlOrBr",
    "YlOrRd",
    "OrRd",
    "PuRd",
    "RdPu",
    "BuPu",
    "GnBu",
    "PuBu",
    "YlGnBu",
    "PuBuGn",
    "BuGn",
    "YlGn",
]

chroms = {
    "chr1",
    "chr10",
    "chr11",
    "chr12",
    "chr13",
    "chr14",
    "chr15",
    "chr16",
    "chr17",
    "chr18",
    "chr19",
    "chr2",
    "chr20",
    "chr21",
    "chr22",
    "chr3",
    "chr4",
    "chr5",
    "chr6",
    "chr7",
    "chr8",
    "chr9",
    "chrX",
    "chrY",
    "1",
    "10",
    "11",
    "12",
    "13",
    "14",
    "15",
    "16",
    "17",
    "18",
    "19",
    "2",
    "20",
    "21",
    "22",
    "3",
    "4",
    "5",
    "6",
    "7",
    "8",
    "9",
    "X",
    "Y",
}


def fromGTF2BED(gtfname, bedname, gtftype="geneAnnot"):
    """
    transforms a  gtf file into a bed file

    Args:
    ----
      gtfname: filepath to gtf file
      bedname: filepath to beddfile
      gtftype: only geneAnnot for now

    Returns:
    --------
      newbed: the bedfile as a pandas.df

    """
    if gtftype == "geneAnnot":
        gtf = pd.read_csv(
            gtfname,
            sep="\t",
            header=0,
            names=[
                "chr",
                "val",
                "type",
                "start",
                "stop",
                "dot",
                "strand",
                "loc",
                "name",
            ],
        )
        gtf["name"] = [
            i.split('gene_id "')[-1].split('"; trans')[0] for i in gtf["name"]
        ]
        prevname = ""
        newbed = {"chr": [], "start": [], "end": [], "gene": []}
        for i, val in gtf.iterrows():
            h.showcount(i, len(gtf))
            if val["name"] == prevname:
                newbed["end"][-1] = val["stop"]
            else:
                newbed["chr"].append(val["chr"])
                newbed["start"].append(val["start"])
                newbed["end"].append(val["stop"])
                newbed["gene"].append(val["name"])
            prevname = val["name"]
        newbed = pd.DataFrame(newbed)
        newbed = newbed[~newbed.chr.str.contains("_fix")]
        newbed.to_csv(bedname + ".bed", sep="\t", index=None)
        newbed.to_csv(bedname + "_genes.bed", sep="\t", index=None)
        return newbed


def getBamDate(bams, split="-", order="des", unknown="U"):
    """
    from bam files (could be in a google bucket) returns their likely sequencing date if available in the header

    Args:
    -----
      bams: the bams file|bucket paths
      split: the splitter in the output date
      unknown: maybe the some dates can't be found the program will output unknown for them
      order: if 'asc', do d,m,y else do y,m,d

    Returns:
    -------
      a list of likely dates or [unknown]s
    """
    DTs = []
    for i, bam in enumerate(tqdm(bams)):
        data = os.popen(
            "export GCS_OAUTH_TOKEN=`gcloud auth application-default print-access-token`\
    && samtools view -H "
            + bam
            + ' | grep "^@RG"'
        )
        if data == signal.SIGINT:
            print("Awakened")
            break
        else:
            res = data.read()
            dt = re.findall("(?<=\tDT:).+?\t", res)
        if len(dt) > 1:
            arr = np.array(dt[0].split("T")[0].split(split)).astype(int)
            for val in dt[1:]:
                arr = np.vstack(
                    (arr, np.array(val.split("T")[0].split(split)).astype(int))
                )
            arr = arr.T
            i = (
                arr[0] * 365 + arr[1] * 31 + arr[2]
                if order == "asc"
                else arr[2] * 365 + arr[1] * 31 + arr[0]
            )
            DTs.append(dt[np.argsort(i)[0]].split("T")[0])
        elif len(dt) == 1:
            DTs.append(dt[0].split("T")[0])
        else:
            DTs.append(unknown)
    return DTs


async def indexBams(bams=None, bucketpath=None, cores=4):
    """
    given a bucket path, will index all .bam files without an associated index and return their paths
    """
    if bams is None:
        if bucketpath is None:
            raise ValueError("need one of bams or bucketpath")
        files = gcp.lsFiles([bucketpath])
        bams = [val for val in files if ".bam" in val[-4:]]
        unindexed = [
            val
            for val in bams
            if val[:-4] + ".bai" not in files and val[:4] + ".bam.bai" not in files
        ]
        print("found " + str(len(unindexed)) + " files to reindex")
    else:
        unindexed = bams
    h.parrun(
        [
            "export GCS_OAUTH_TOKEN=`gcloud auth application-default print-access-token` && samtools index "
            + val
            for val in unindexed
        ],
        cores,
    )
    return {val: val[:-4] + ".bam.bai" for val in unindexed}


def dropWeirdChromosomes(bedfile, keep=[], skip=0):
    """
    given a bedfile path, removes chromosomes that are not one of chroms

    Args:
    ----
      bedfile: str the filepath to the bedfile
      keep: list[str] of additional chromosomes to keep
    """
    if skip >= 20:
        raise ValueError("too many header lines!")
    try:
        bed = pd.read_csv(bedfile, sep="\t", header=None, skiprows=skip)
    except ParserError:
        dropWeirdChromosomes(bedfile, keep, skip + 1)
        return
    except EmptyDataError:
        print("empty bed")
        return
    initlen = len(bed)
    if initlen == 0:
        print("empty bed")
        return
    bed = bed[bed[0].isin(chroms | set(keep))]
    if len(bed) < skip and skip > 5:
        raise ValueError("too many header lines!")
    print("found " + str(skip) + " header line... removing")
    if len(bed) != initlen:
        print("removed " + str(initlen - len(bed)) + " lines")
    bed.to_csv(bedfile, sep="\t", header=None, index=None)


def extractPairedSingleEndFrom(folder, sep="-", namepos=2):
    """
    given a folder, find fastq files and sorts paired and single end based on the R1/R2 patterns

    Args:
    -----
      folder: the folder where the fastqs are
      sep: the separator in filename
      namepos: the location of the name in this separated list of name from filepath

    Returns:
    -------
      list of filepath to single end files
      df with R1 and R2 filepath
    """
    single = []
    paired = {}
    for val in os.listdir(folder):
        if ".fastq" in val or ".fq" in val:
            if "R1" in val:
                name = val.split(sep)[namepos]
                paired[name] = {"R1": val}
            elif "R2" in val:
                name = val.split(sep)[namepos]
                paired[name].update({"R2": val})
            else:
                single.append(val)
    return single, pd.DataFrame(paired)


def findReplicatesBams(folder, sep="-", namings="-r([0-9])", namepos=2):
    """
    creates a dict of name and replicate files given a regexp namging scheme
    """
    rep = {}
    for val in os.listdir(folder):
        if val[-4:] == ".bam":
            match = re.search(namings, val)
            if match:
                name = val.split(sep)[namepos]
                if name in rep:
                    rep[name].append(val)
                else:
                    rep[name] = [val]

    return rep


def singleEnd(
    singlend,
    folder="data/seqs/",
    numthreads=8,
    peaksFolder="peaks/",
    ismapped=False,
    mappedFolder="mapped/",
    refFolder="data/reference/index",
):
    """
    run the singleEnd pipeline
    for alignment etc, one can use pysam ready made implementation of samtools
    """
    print(
        "you need to have bowtie2 installed: http://bowtie-bio.sourceforge.net/bowtie2/index.shtml"
    )
    for val in singlend:
        out1 = folder + mappedFolder + val.split(".")[0] + ".mapped.sam"
        if not ismapped:
            in1 = folder + val
            os.system(
                "bowtie2 -x "
                + refFolder
                + " --threads "
                + str(numthreads)
                + " -t -k 1 --very-sensitive -U "
                + in1
                + " -S "
                + out1
            )
        out2 = folder + peaksFolder + val.split(".")[0]
        print(out1)
        os.system("macs2 callpeak -f SAM -t " + out1 + " --outdir " + out2)
        # it can take many TB so better delete


def pairedEnd(
    pairedend,
    folder="",
    numthreads=8,
    peaksFolder="peaks/",
    ismapped=False,
    mappedFolder="mapped/",
    refFolder="data/reference/index",
):
    """
    # run the paired end pipeline
    """
    print(
        "you need to have bowtie2 installed: http://bowtie-bio.sourceforge.net/bowtie2/index.shtml"
    )
    for _, val in pairedend.items():
        out1 = folder + mappedFolder + val[0].split(".")[0] + ".mapped.sam"
        in1 = folder + val[0]
        in2 = folder + val[1]
        os.system(
            "bowtie2 -x "
            + refFolder
            + " --threads "
            + str(numthreads)
            + " -t -k 1 \
    --very-sensitive -1 "
            + in1
            + " -2 "
            + in2
            + " - S "
            + out1
        )
        out2 = folder + peaksFolder + val[0].split(".")[0]
        print(out1)
        changefrom = out1
        changeto = out1[:-4] + ".bam"
        os.system("samtools view -b " + changefrom + " -o " + changeto)
        os.system(
            "macs2 callpeak --format 'BAMPE' --treatment "
            + changeto
            + " --outdir "
            + out2
        )
        # it can take many TB so better delete


async def mergeBams(rep):
    """
    uses samtools to merge a set of replicates considered into one file
    """
    in1 = ""
    for i, val in rep.items():
        out1 = i + ".merged.bam"
        for bam in val:
            in1 += " " + bam
        os.system("samtools merge " + out1 + in1)


def compare_gcloud_vcfs_overlap_methods(vcfs_met1_path, vcfs_met2_path):
    for i, j in zip(vcfs_met1_path, vcfs_met2_path):
        compare_gcloud_vcf_overlap(i, j)


def compare_gcloud_vcf_overlap(vcf1, vcf2, cols=["chr", "start", ".", "ref", "alt"]):
    import subprocess

    name1 = vcf1.split("/")[-1].split(".")[0] + "_1" + ".tsv"
    cmd1 = "gsutil cat " + vcf1 + " | gunzip | cut -f -5 > " + name1
    name2 = vcf2.split("/")[-1].split(".")[0] + "_2" + ".tsv"
    cmd2 = "gsutil cat " + vcf2 + " | gunzip | cut -f -5 > " + name2
    try:
        subprocess.run(
            cmd1,
            shell=True,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        subprocess.run(
            cmd2,
            shell=True,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    except subprocess.CalledProcessError as e:
        print(e.stderr)
        raise e
    val2 = pd.read_csv(name2, sep="\t", comment="#", names=cols)
    val1 = pd.read_csv(name1, sep="\t", comment="#", names=cols)
    val1["loc"] = (
        val1["chr"].astype(str)
        + ":"
        + val1["start"].astype(str)
        + ":"
        + val1["alt"].astype(str)
    )
    val2["loc"] = (
        val2["chr"].astype(str)
        + ":"
        + val2["start"].astype(str)
        + ":"
        + val2["alt"].astype(str)
    )
    print("length of vcf1:" + str(len(val1)))
    print("length of vcf2:" + str(len(val2)))
    print("overlap: " + str(len(set(val1["loc"]).intersection(val2["loc"]))))
    return val1, val2



from . import name_mapper, arxspan_to_ccle, ccle_to_arxspan, latest_ccle_names, csv_url
import cell_line_mapper
import pytest
import pandas as pd
import requests
from io import StringIO
from collections import defaultdict
import csv
##################################################################################
### TEST ARXSPAN_TO_CCLE
#just the first three broadid's in the doc

def test_real_data_fetch_check():
    return_dict = defaultdict(set)

    mapping_request = StringIO(requests.get(csv_url).text)
    df = pd.read_csv(mapping_request)
    assert list(df) == ['ccle_name', 'canonical_ccle_name', 'broad_id']


##################################################################################

@pytest.fixture()
def fake_mapping_csv(monkeypatch):
    def mock_read_file_to_dict(key_name, value_name):
        return_dict = defaultdict(str)

        with open('test-data.csv', mode='r') as csvfile:
            map_reader = csv.DictReader(csvfile)
            for rows in map_reader:
                if return_dict[rows[key_name]] == "":
                    return_dict[rows[key_name]] = rows[value_name]
                else:
                    return_dict[rows[key_name]]+=str(", "+rows[value_name])
        print(return_dict)
        return return_dict

    monkeypatch.setattr(cell_line_mapper, 'read_file_to_dict', mock_read_file_to_dict)


##################################################################################
### TEST ARXSPAN_TO_CCLE
# just the first three broadid's in the doc
def test_arxspan_to_ccle_first_three_rows(fake_mapping_csv):
    assert arxspan_to_ccle(["1", "2", "3"]) == ["A", "B", "C"]


#  1.  unmappable ID and ignore_problems == false
def test_arxspan_to_ccle_unmappable_ID_ignore_problems_false(fake_mapping_csv):
    with pytest.raises(KeyError) as excinfo:
        arxspan_to_ccle(["1", "madeupfakename"])
    assert "canonical_ccle_name could not be found for the following broad_id:  madeupfakename" in str(excinfo.value)


#  2.  unmappable ID and ignore_problems == true
def test_arxspan_to_ccle_unmappable_ID_ignore_problems_true(fake_mapping_csv):
    assert arxspan_to_ccle(["1", "madeupfakename"], True) == ["A", None]


#  2.5.  arxspan id has multiple ccle names
def test_arxspan_to_ccle_one_key_many_values(fake_mapping_csv):
    assert arxspan_to_ccle(["7"], True, False) == ["D, F"]


#  3.  unique elements in arxspan_ids do not map to unique ccle names and check_unique_mapping == false
def test_arxspan_to_ccle_nonunique_mapping_check_unique_mapping_false(fake_mapping_csv):
    assert arxspan_to_ccle(["4", "5"], True, False) == ["D", "D"]


#  4.  unique elements in arxspan_ids do not map to unique ccle names and check_unique_mapping == true
def test_arxspan_to_ccle_nonunique_mapping_check_unique_mapping_true(fake_mapping_csv):
    with pytest.raises(RuntimeError) as excinfo:
        arxspan_to_ccle(["4", "5"], True, True)
    assert 'Mappings are not unique' in str(excinfo.value)


##################################################################################
### TEST CCLE_ARXSPAN
#  0.  Just the first three rows
def test_ccle_to_arxspan_first_three_rows(fake_mapping_csv):
    assert ccle_to_arxspan(["a", "B", "c"]) == ["1", "2", "3"]


#  1.  unmappable ccle name and ignore_problems == false
def test_ccle_to_arxspan_unmappable_ID_ignore_problems_false(fake_mapping_csv):
    with pytest.raises(KeyError) as excinfo:
        ccle_to_arxspan(["a", "madeupfakename", "idk"], False)
        assert "broad_id could not be found for the following canonical_ccle_name:  madeupfakename" in str(excinfo.value)


#  2.  unmappable ccle name and ignore_problems == true
def test_ccle_to_arxspan_unmappable_ID_ignore_problems_true(fake_mapping_csv):
    assert ccle_to_arxspan(["a", "madeupfakename"], True) == ["1", None]


#  2.5.  ccle name has multiple arxspan_ids
def test_ccle_to_arxspan_one_key_many_values(fake_mapping_csv):
    assert ccle_to_arxspan(["d"], True, False) == ["4, 5, 7"]


#  3.  unique ccle names do not map to unique arxspan_ids and check_unique_mapping == false
def test_ccle_to_arxspan_nonunique_mapping_check_unique_mapping_false(fake_mapping_csv):
    assert ccle_to_arxspan(["e"], True, False) == ["6, 7"]


#  4.  unique ccle names do not map to unique arxspan_ids and check_unique_mapping == true
def test_ccle_to_arxspan_nonunique_mapping_check_unique_mapping_true(fake_mapping_csv):
    with pytest.raises(RuntimeError) as excinfo:
        a = ccle_to_arxspan(["e"], True, True)
    assert 'Mappings are not unique' in str(excinfo.value)


##################################################################################
### TEST LATEST_CCLE_NAMES

#  0.  Just a few samples that have different names
def test_latest_ccle_names_first_three_rows(fake_mapping_csv):
    assert latest_ccle_names(["a", "B", "c"]) == ["A", "B", "C"]


#  1.  unmappable ccle name and ignore_problems == false
def test_latest_ccle_names_unmappable_ignore_problems_false(fake_mapping_csv):
    with pytest.raises(KeyError) as excinfo:
        latest_ccle_names(["madeupfakename"], False)
    assert "canonical_ccle_name could not be found for the following ccle_name:  madeupfakename" in str(excinfo.value)


#  2.  unmappable ccle name and ignore_problems == true
def test_latest_ccle_names_unmappable_ignore_problems_true(fake_mapping_csv):
    assert latest_ccle_names(["madeupfakename", "a"], True) == [None, "A"]


#    2.5  old ccle name maps to multiple latest names
def test_latest_ccle_names_one_key_many_values(fake_mapping_csv):
    assert latest_ccle_names(["e"], True, False) == ["E, F"]

def test_latest_ccle_names_one_key_many_values_check_unique_mapping_true(fake_mapping_csv):
    with pytest.raises(RuntimeError) as excinfo:
        latest_ccle_names(["e"], True, True)
    assert 'Mappings are not unique' in str(excinfo.value)

#  4.  unique ccle names do not map to unique latest names and check_unique_mapping == true
def test_latest_ccle_names_nonunique_mapping_check_unique_mapping_true(fake_mapping_csv):
    with pytest.raises(RuntimeError) as excinfo:
        latest_ccle_names(["d", "dd", "ddd"], True, True)
    assert 'Mappings are not unique' in str(excinfo.value)



#!/usr/bin/env python

import ast
import re
from setuptools import setup, find_packages

_version_re = re.compile(r'__version__\s*=\s*(.*)')

with open('cell_line_mapper/__init__.py', 'rt') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read()).group(1)))

setup(name='cell_line_mapper',
      version=version,
      description='Functions for mapping between cell line identifiers',
      author='Phoebe Moh',
      author_email='pmoh@broadinstitute.org',
      install_requires=['pandas', 'requests'],
      packages=find_packages()
     )




import os
import pandas as pd
import numpy as np
from genepy.utils import helper as h
from genepy.utils import plot
from genepy.epigenetics.chipseq import *
import seaborn as sns
import pyBigWig
import matplotlib.pyplot as plt
from scipy.optimize import minimize
from scipy.special import factorial
import warnings
import itertools


def findpeakpath(folder, proteiname):
    """
    given a folder of bigwigs and a protein name, finds the right bigwig
    """
    res = None
    for val in os.listdir(folder):
        if str(proteiname) in val:
            if res:
                raise ValueError('more than 1 bigwig file found')
            res = val
    if res:
        return res
    raise ValueError('no bigwig file found')


def findBestPeak(presence):
    """
    given a list of -sets of peak locations for each replicate- will return the best replicate given a simple metric
    """
    tot = []
    for ind, el in enumerate(presence):
        val = len(el)
        pres = [x for j, x in enumerate(presence) if j != ind]
        for jnd in range(1, len(pres)+1):
            for comb in itertools.combinations(pres, jnd):
                ov = el
                for knd in range(jnd):
                    ov = ov & comb[knd]
                val += len(ov)*(jnd+1)
        tot.append(val)
    return np.argsort(tot)[::-1]


def mergeReplicatePeaks(peaks, bigwigfolder, markedasbad=None, window=100,
                                                sampling=3000, mincov=4, doPlot=True, cov={}, minKL=8, use='max',
                                                MINOVERLAP=0.3, lookeverywhere=True, only='', saveloc=''):
    """
    /!/ should only be passed peaks with at least one good replicate
    for each TFpeaksets,
    1. find the replicate that have the most peaks
    2. correlate peaks and get in highest correlation order with the replicate found in 1
    3. find overlap of both and get size of second replicate
    4. if small(er)-> use only to increase statistics
        1. if a lot of uncalled peaks in replicate 2 at replicate 1 peaks (flag for mergebam)
    5. if similar size -> get only intersect
        2. add to intersect, find uncalled peaks in both replicates which are called in the other
    6. repeat for all replicates
    -------------------------
    if full overlap of one of the peak replicate, only use the overlapped one to increase confidence on peak
    if >80% average non overlap,
        print warning and percentage of overlap

    if <20% average non overlap,
        take the overlap and increase confidence and avg logfold

    if one is <20%:
        if other <40% average non overlap,
        take the overlap and increase confidence and avg logfold
        else
        take

    gets the max cov at the genomic window and if above some threshold, accepts the peak.

    extend peak by X bp if no TSS
    remove TSS from peaks


        create a new data frame containing merged peak size, reassembled peak data (p value etc..) and
        a the value for presence of each TF listed in previous df
        ------------------------------------

        args:
        ----
        peaks: df[bed-like] all the peaks into the sameBam with a column containing the 'name'
        being the id of the sample, the 'replicate' number of this sample, the 'tf' chiped here
        bamfolder: str, foldername
        avgCov: dict(filename:int) a dict where for each bam filename is given an averageCoverage
        if use=='max':
                window:
                mincov:

        if use=='max':


        returns:
        -------
        mergedpeaks: dict{df-peakslike}
        bamtomerge: [[bam1,bam2]]
    """
    def col_nan_scatter(x, y, **kwargs):
        df = pd.DataFrame({'x': x[:], 'y': y[:]})
        df = df[df.sum(0) != 0]
        x = df['x']
        y = df['y']
        plt.gca()
        plt.scatter(x, y)

    def col_nan_kde_histo(x, **kwargs):
        df = pd.DataFrame({'x': x[:]})
        df = df[df['x'] != 0]
        x = df['x']
        plt.gca()
        sns.kdeplot(x)
    print("/!/ should only be passed peaks with at least one good replicate")
    # for a df containing a set of peaks in bed format and an additional column of different TF
    tfs = list(set(peaks['tf']))
    totpeaknumber = 0
    mergedpeaksdict = {}
    remove = []
    tomergebam = []
    ratiosofunique = {}
    h.createFoldersFor(saveloc)
    f = open(saveloc+'results.txt', 'w')
    warnings.simplefilter("ignore")
    for tf in tfs:
        if only and tf != only:
            continue
        cpeaks = peaks[peaks.tf == tf]
        print('_____________________________________________________')
        f.write('_____________________________________________________' + '\n')
        if len(set(cpeaks['replicate'])) == 1:
            if cpeaks.name.tolist()[0] in markedasbad:
                print('the only replicate is considered bad!')
                f.write('the only replicate is considered bad!'+"\n")
                print('wrong TF: '+tf)
                f.write('wrong TF: '+tf+"\n")
                mergedpeaksdict.update({tf: cpeaks})
                remove.append(tf)
                continue
            print("we only have one replicate for " + tf + " .. pass")
            f.write("we only have one replicate for " + tf + " .. pass"+"\n")
            mergedpeaksdict.update({tf: cpeaks})
            continue
        print("merging " + tf + " peaks")
        f.write("merging " + tf + " peaks"+"\n")
        merged = simpleMergePeaks(cpeaks, window=window, maxp=False)
        merged_bed = merged[merged.columns[8:]]
        finalpeaks = merged[merged.columns[:8]]
        print('--> finish first overlaps lookup')
        f.write('--> finish first overlaps lookup'+"\n")
        # flag when  biggest is <1000 peaks
        if len(finalpeaks) < 1000:
            print('!TF has less than 1000 PEAKS!')
            f.write('!TF has less than 1000 PEAKS!'+"\n")
        # for each TF (replicates), compute number of peaks
        peakmatrix = merged_bed.values.astype(bool)

        presence = []
        for peakpres in peakmatrix.T:  # https://github.com/tctianchi/pyvenn
            presence.append(set([i for i, val in enumerate(peakpres) if val == 1]))
        # compute overlap matrix (venn?)
        if peakmatrix.shape[1] < 7 and doPlot:
            plot.venn(presence, [i+'_BAD' if i.split('-')[0]
                                                in markedasbad else i for i in merged_bed.columns], title=tf+"_before_venn", folder=saveloc)
            plt.show()
        else:
            print('too many replicates for Venn: '+str(peakmatrix.shape[1]))
            f.write('too many replicates for Venn: '+str(peakmatrix.shape[1])+"\n")
        if doPlot:
            fig = sns.pairplot(merged_bed, corner=True, diag_kind="kde",
                                                 kind="reg", plot_kws={"scatter_kws": {"alpha": .05}})
            #fig = fig.map_upper(col_nan_scatter)
            #fig = fig.map_upper(col_nan_kde_histo)
            plt.suptitle("correlation of peaks in each replicate", y=1.08)
            if saveloc:
                fig.savefig(saveloc+tf+"_before_pairplot.pdf")
            plt.show()
            for i, val in enumerate(merged_bed):
                unique_inval = np.logical_and(
                    np.delete(peakmatrix, i, axis=1).sum(1).astype(bool) == 0, peakmatrix[:, i])
                sns.kdeplot(merged_bed[val][unique_inval], legend=True).set(xlim=(0, None))
            plt.title("distribution of unique peaks in each replicate")
            if saveloc:
                plt.savefig(saveloc+tf+"_before_unique_kdeplot.pdf")
            plt.show()

        bigwigs = os.listdir(bigwigfolder)

        foundgood = False
        sort = findBestPeak(presence)
        for ib, sb in enumerate(sort):
            if merged_bed.columns[sb].split('-')[0] not in markedasbad:
                foundgood = True
                break
        if not foundgood:
            print('no peaks were good enough quality')
            f.write('no peaks were good enough quality'+"\n")
            print('bad TF: '+tf)
            f.write('bad TF: '+tf+"\n")
            remove.append(tf)
            ib = 0
        # distplot
        # correlation plot

        biggest_ind = sort[ib]
        peakmatrix = peakmatrix.T
        biggest = merged_bed.columns[biggest_ind]
        print('-> main rep is: '+str(biggest))
        f.write('-> main rep is: '+str(biggest)+'\n')
        tot = peakmatrix[biggest_ind].copy().astype(int)
        # starts with highest similarity and go descending
        j = 0
        recovered = 0
        additionalpeaksinbig = np.array([])
        for i, val in enumerate(sort):
            if i == ib:
                continue
            j += 1
            # if avg non overlap > 60%, and first, and none small flag TF as unreliable.
            overlap = len(presence[biggest_ind] & presence[val]
                                        ) / len(presence[biggest_ind])
            peakname = merged_bed.columns[val]
            print('- '+peakname)
            f.write('- '+peakname+'\n')
            print('  overlap: ' + str(overlap*100)+"%")
            f.write('  overlap: ' + str(overlap*100)+"%"+'\n')
            if overlap < MINOVERLAP:
                smallsupport = len(presence[biggest_ind] &
                                                     presence[val]) / len(presence[val])
                print(' --> not enough overlap')
                f.write(' --> not enough overlap'+'\n')
                if smallsupport < MINOVERLAP:
                    # if the secondary does not have itself the required support
                    if j == 1 and merged_bed.columns[val].split('-')[0] not in markedasbad:
                        print("  Wrong TF: "+tf)
                        f.write("  Wrong TF: "+tf+'\n')
                        remove.append(tf)
                        break
                    # if not first, throw the other replicate and continue
                    print("  not using this replicate from the peakmatrix")
                    f.write("  not using this replicate from the peakmatrix"+'\n')
                    continue
            if lookeverywhere:
                tolookfor = peakmatrix[val] == 0
            else:
                tolookfor = np.logical_and(peakmatrix[biggest_ind], peakmatrix[val] == 0)
            # ones that we have in the Primary but not in the secondary
            additionalpeaksinsec = findAdditionalPeaks(finalpeaks, tolookfor, bigwigfolder + findpeakpath(
                bigwigfolder, peakname), sampling=sampling, mincov=mincov, window=window, minKL=minKL, use=use)
            if len(additionalpeaksinsec[additionalpeaksinsec > 0]) > 0:
                sns.kdeplot(additionalpeaksinsec[additionalpeaksinsec > 0],
                                        label=peakname, legend=True).set(xlim=(0, None))
                print('  min,max from newly found peaks: ' +
                            str((additionalpeaksinsec[additionalpeaksinsec > 0].min(), additionalpeaksinsec[additionalpeaksinsec > 0].max())))
                f.write('  min,max from newly found peaks: '+str((additionalpeaksinsec[additionalpeaksinsec > 0].min(
                ), additionalpeaksinsec[additionalpeaksinsec > 0].max()))+'\n')
            # for testing purposes mainly
            finalpeaks[additionalpeaksinsec.astype(bool)].to_csv(
                'additionalpeaksinsec_mp'+merged_bed.columns[val]+'.bed', sep='\t', index=None, header=False)
            peakmatrix[val] = np.logical_or(
                peakmatrix[val], additionalpeaksinsec.astype(bool))
            overlap = np.sum(np.logical_and(
                peakmatrix[val], peakmatrix[biggest_ind]))/np.sum(peakmatrix[biggest_ind])
            if overlap < MINOVERLAP:
                newsmalloverlap = np.sum(np.logical_and(
                    peakmatrix[val], peakmatrix[biggest_ind]))/np.sum(peakmatrix[val])
                print("  we did not had enough initial overlap.")
                f.write("  we did not had enough initial overlap."+'\n')
                if newsmalloverlap < MINOVERLAP:
                    if merged_bed.columns[val].split('-')[0] in markedasbad:
                        print('  replicate ' +
                                    merged_bed.columns[val] + ' was too bad and had not enough overlap')
                        f.write('  replicate ' +
                                        merged_bed.columns[val] + ' was too bad and had not enough overlap'+'\n')
                        continue
                    elif h.askif("we have two good quality peaks that don't merge well at all: "+merged_bed.columns[val] +
                                                                                        " and " + merged_bed.columns[biggest_ind] + " can the first one be removed?:\n  \
                            overlap: "+str(overlap*100)+'%\n  smalloverlap: '+str(smalloverlap*100)+'%\n  new smalloverlap: '+str(newsmalloverlap*100)+"%"):
                        continue
                    else:
                        print("  enough from small overlaps")
                        f.write("  enough from small overlaps"+'\n')
            print(' --> enough overlap')
            f.write(' --> enough overlap'+'\n')
            recovered += np.sum(additionalpeaksinsec.astype(bool))
            if merged_bed.columns[val].split('-')[0] not in markedasbad:
                tot += peakmatrix[val].astype(int)
            # ones that we have in the Primary but not in the secondary
            if not lookeverywhere or len(additionalpeaksinbig) == 0:
                tolookfor = peakmatrix[biggest_ind] == 0 if lookeverywhere else np.logical_and(
                    peakmatrix[biggest_ind] == 0, peakmatrix[val])
                additionalpeaksinbig = findAdditionalPeaks(finalpeaks, tolookfor, bigwigfolder + findpeakpath(
                    bigwigfolder, biggest), sampling=sampling, mincov=mincov, window=window, minKL=minKL, use=use)
                if len(additionalpeaksinbig[additionalpeaksinbig > 0]) > 0:
                    sns.kdeplot(additionalpeaksinbig[additionalpeaksinbig > 0],
                                            label=biggest, legend=True).set(xlim=(0, None))
                    print('  min,max from newly found peaks: ' +
                                str((additionalpeaksinbig[additionalpeaksinbig > 0].min(), additionalpeaksinbig[additionalpeaksinbig > 0].max())))
                    f.write('  min,max from newly found peaks: '+str((additionalpeaksinbig[additionalpeaksinbig > 0].min(
                    ), additionalpeaksinbig[additionalpeaksinbig > 0].max()))+'\n')

                peakmatrix[biggest_ind] = np.logical_or(
                    peakmatrix[biggest_ind], additionalpeaksinbig)
                tot += additionalpeaksinbig.astype(bool).astype(int)
                recovered += np.sum(additionalpeaksinbig.astype(bool))
            print('  we have recovered ' + str(recovered)+' peaks, equal to ' + str(100*recovered/np.sum(peakmatrix[biggest_ind])) +
                                                        '% of the peaks in main replicate')
            f.write('  we have recovered ' + str(recovered)+' peaks, equal to ' + str(100*recovered/np.sum(peakmatrix[biggest_ind])) +
                                                        '% of the peaks in main replicate'+'\n')
            if overlap < (MINOVERLAP+0.2)/1.2:
                # we recompute to see if the overlap changed
                newoverlap = np.sum(np.logical_and(
                    peakmatrix[val], peakmatrix[biggest_ind]))/np.sum(peakmatrix[biggest_ind])
                smalloverlap = np.sum(np.logical_and(
                    peakmatrix[val], peakmatrix[biggest_ind]))/np.sum(peakmatrix[val])
                if newoverlap < (MINOVERLAP+0.2)/1.2:
                    if smalloverlap < (2+MINOVERLAP)/3:
                        print("  not enough overlap to advice to merge the bams.\n  oldnew overlap: "+str(overlap*100)+'%\n  \
                            new overlap: '+str(newoverlap*100)+"%")
                        f.write("  not enough overlap to advice to merge the bams.\n  oldnew overlap: "+str(overlap*100)+'%\n  \
                            new overlap: '+str(newoverlap*100)+"%"+'\n')
                        continue
                    else:
                        print('  enough from small overlap to advice to merge the peaks')
                        f.write('  enough from small overlap to advice to merge the peaks'+'\n')
            tomergebam.append([biggest, peakname])
            #the quality is good enough in the end we can pop from the list if it exists
            if tf in remove:
                remove.remove(tf)
        plt.title('distribution of new found peaks')
        if saveloc:
            plt.savefig(saveloc+tf+"_new_found_peaks_kdeplot.pdf")
        plt.show()
        # new distplot
        # new correlation plot
        ratiosofunique[tf] = len(np.argwhere(
            peakmatrix.sum(0) == 1))/peakmatrix.shape[1]
        if doPlot:
            sns.pairplot(merged_bed, corner=True, diag_kind="kde",
                                     kind="reg", plot_kws={"scatter_kws": {"alpha": .05}})
            #fig = fig.map_upper(col_nan_scatter)
            #fig = fig.map_upper(col_nan_kde_histo)
            plt.suptitle("correlation and distribution of peaks after recovery", y=1.08)
            if saveloc:
                plt.savefig(saveloc+tf+"_after_pairplot.pdf")
            plt.show()
            for i, val in enumerate(merged_bed):
                unique_inval = np.logical_and(
                    np.delete(peakmatrix, i, axis=0).sum(0).astype(bool) == 0, peakmatrix[i])
                sns.kdeplot(merged_bed[val][unique_inval], legend=True).set(xlim=(0, None))
            plt.title("distribution of unique peaks in each replicate after recovery")
            if saveloc:
                plt.savefig(saveloc+tf+"_after_unique_kdeplot.pdf")
            plt.show()
        if len(peakmatrix.shape) > 1 and doPlot:
            if peakmatrix.shape[0] < 7:
                presence = []
                for peakpres in peakmatrix:  # https://github.com/tctianchi/pyvenn
                    presence.append(set([i for i, val in enumerate(peakpres) if val == 1]))
                title = tf + '_recovered (TOREMOVE)' if tf in remove else tf+'_recovered'
                plot.venn(presence, [i+'_BAD' if i.split('-')[0]
                                                                        in markedasbad else i for i in merged_bed.columns], title=title, folder=saveloc)
                plt.show()
            else:
                print('too many replicates for Venn')
                f.write('(too many replicates for Venn)'+'\n')
            finalpeaks = finalpeaks[np.logical_or(tot > 1, peakmatrix[biggest_ind])]
        finalpeaks['name'] = biggest
        finalpeaks['tf'] = tf
        mergedpeaksdict.update({tf: finalpeaks})
        print(str((tf, len(finalpeaks))))
        f.write(str((tf, len(finalpeaks)))+'\n')
    mergedpeak = pd.concat(
        [peaks for _, peaks in mergedpeaksdict.items()]).reset_index(drop=True)
    if doPlot:
        df = pd.DataFrame(data=ratiosofunique, index=['percentage of unique'])
        df['proteins'] = df.index
        fig = sns.barplot(data=df)
        plt.xticks(rotation=60, ha='right')
        plt.title("ratios of unique in replicates across experiments")
        if saveloc:
            plt.savefig(saveloc+"All_ratios_unique.pdf")
        plt.show()
    f.close()
    mergedpeak['name'] = mergedpeak.tf
    return mergedpeak, tomergebam, remove, ratiosofunique


def findAdditionalPeaks(peaks, tolookfor, filepath, sampling=1000, mincov=4,
                                                window=100, cov={}, minKL=8, use='max'):

    """
    findAdditionalPeaks: for all peaks in A and/or B find in coverage file if zone has relative cov
    of more than thresh then add to peak
    if B is small and > 20% of peaks in A are found back, increase confidence and
    flag for mergeBams
    if < 20% don't flag for merge bam
    f B is big and now mean non overlap < 40%, take union and flag for mergeBam else, throw B.

    Args:
    -----
        peaks
        tolookfor
        filepath
        sampling
        mincov
        window
        cov
        minKL
        use
    returns:
    -------
        np.array(bool) for each peaks in peakset, returns a binary
    """
    # def poisson(k, lamb, scale): return scale * (lamb**k / factorial(k)) * np.exp(-lamb)

    def KLpoisson(lamb1, lamb2): return lamb1 * \
                        np.log(lamb1 / lamb2) + lamb2 - lamb1

    def poisson(k, lamb): return (lamb**k/factorial(k)) * np.exp(-lamb)

    def negLogLikelihood(params, data): return - \
                        np.sum(np.log(poisson(data, params[0])))

    def poissonFit(data): return float(
        minimize(negLogLikelihood, x0=np.ones(1), args=(data,), method='Powell').x)
    bw = pyBigWig.open(filepath)
    res = np.zeros(len(peaks))
    prevchrom = ''
    lamb = {}
    cov = {}
    #ignore by message
    warnings.filterwarnings("ignore", message="encountered in")
    for i, has in enumerate(tolookfor):
        if has:
            val = peaks.iloc[i]
            if val.chrom not in chroms:
                continue
            if val.chrom != prevchrom:
                if val.chrom not in cov:
                    cov[val.chrom] = bw.stats(str(val.chrom))[0]
                    prevchrom = val.chrom
                    if use == 'poisson':
                        #TODO: compute on INPUT file instead
                        samples = np.zeros(window * sampling)
                        sam = np.random.rand(sampling)
                        sam = sam * (bw.chroms(str(val.chrom))-window)
                        for j, sample in enumerate(sam.astype(int)):
                            samples[j*window:(j + 1)*window] = np.nan_to_num(
                                bw.values(str(val.chrom), sample, sample + window), 0)
                        scale = np.unique(samples)[1]
                        samples = (samples/scale).astype(int)
                        lamb[val.chrom] = (poissonFit(samples), scale)

            start = max([val.start - window, 0])
            end = min(val.end + window, bw.chroms(str(val.chrom)))
            zone = np.nan_to_num(bw.values(str(val.chrom), start, end), 0)
            if use == 'max':
                if max(zone) / cov[val.chrom] > mincov*1.5 or sum(zone) / (cov[val.chrom] * (end - start)) > mincov:
                    res[i] = max(zone) / cov[val.chrom]
            elif use == 'poisson':
                #TODO: compute -log10pvalue
                la = poissonFit((zone/lamb[val.chrom][1]).astype(int))
                kl = KLpoisson(la, lamb[val.chrom][0])
                if kl > minKL:
                    res[i] = max(zone) / cov[val.chrom]  # foldchange from macs3

    return res
